#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2024 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from PySide6.QtCore import Qt
from PySide6.QtGui import QResizeEvent, QKeyEvent
from PySide6.QtWidgets import QApplication, QWidget

from pyrdp.player.BaseTab import BaseTab
from pyrdp.player.PlayerEventHandler import PlayerEventHandler
from pyrdp.player.Replay import Replay, ReplayReader
from pyrdp.player.ReplayBar import ReplayBar
from pyrdp.player.ReplayThread import ReplayThread
from pyrdp.ui import QRemoteDesktop

class ReplayTab(BaseTab):
    """
    Tab that displays a RDP Connection that is being replayed from a file.
    """

    def __init__(self, fileName: str, parent: QWidget):
        """
        :param fileName: name of the file to read.
        :param parent: parent widget.
        """
        self.viewer = QRemoteDesktop(800, 600, parent)
        super().__init__(self.viewer, parent)
        QApplication.instance().aboutToQuit.connect(self.onClose)

        self.fileName = fileName
        self.file = open(self.fileName, "rb")
        self.eventHandler = PlayerEventHandler(self.widget, self.text)

        replay = Replay(self.file)
        self.reader = ReplayReader(replay)
        self.thread = ReplayThread(replay)
        self.thread.eventReached.connect(self.readEvent)
        self.thread.timeUpdated.connect(self.onTimeUpdated)
        self.thread.clearNeeded.connect(self.clear)
        self.thread.start()

        self.controlBar = ReplayBar(replay.duration, replay.referenceTime)
        self.controlBar.play.connect(self.thread.play)
        self.controlBar.pause.connect(self.thread.pause)
        self.controlBar.seek.connect(self.thread.seek)
        self.controlBar.speedChanged.connect(self.thread.setSpeed)
        self.controlBar.scaleCheckbox.stateChanged.connect(self.setScaleToWindow)
        self.controlBar.button.setDefault(True)

        self.tabLayout.insertWidget(0, self.controlBar)

    def play(self):
        self.controlBar.button.setPlaying(True)
        self.controlBar.play.emit()

    def readEvent(self, position: int):
        """
        Read an event from the file at the given position.
        :param position: the position of the event in the file.
        """
        event = self.reader.readEvent(position)
        self.eventHandler.onPDUReceived(event)

    def onTimeUpdated(self, currentTime: float):
        """
        Called everytime the thread ticks.
        :param currentTime: the current time.
        """
        self.controlBar.timeSlider.blockSignals(True)
        self.controlBar.timeSlider.setValue(int(currentTime * 1000))
        self.controlBar.timeSlider.blockSignals(False)
        self.controlBar.onTimeChanged(currentTime)

    def clear(self):
        """
        Clear the UI.
        """
        self.viewer.clear()
        self.text.setText("")

    def onClose(self):
        self.thread.close()
        self.thread.wait()
        self.eventHandler.cleanup()

    def setScaleToWindow(self, status: int):
        """
        Called when the scale to window checkbox is checked or unchecked, refresh
        the scaling calculation.
        :param status: state of the checkbox
        """
        self.widget.setScaleToWindow(status)
        self.parentResized(None)

    def parentResized(self, event: QResizeEvent):
        """
        Called when the main PyRDP window is resized to allow to scale the current
        RDP session being displayed.
        :param event: The event of the parent that has been resized
        """
        newScale = self.scrollViewer.viewport().height() / self.widget.sessionHeight
        self.widget.scale(newScale)

    def keyPressEvent(self, event: QKeyEvent):
        """
        Called every time a key is pressed on the Replay tab
        :param event: the key event holding key info
        """
        if event.key() == Qt.Key.Key_Space:
            self.controlBar.button.setPlaying(not self.controlBar.button.playing)
            self.controlBar.onButtonClicked()



#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
GDI Cache Management Layer.
"""

from typing import List
from PySide6.QtCore import QSize
from PySide6.QtGui import QBrush, QImage, QBitmap

from pyrdp.parser.rdp.orders.common import Glyph


class BitmapCache:
    """Bitmap cache. Structured like this: [cache id][cache index] = (image, buffer)"""

    def __init__(self, persist=False):
        self.caches = {}
        self.persist = persist
        if persist:
            raise Exception('Persistent cache is not supported yet.')

    def has(self, cid: int, idx: int) -> bool:
        """
        Check whether a cache contains an entry.

        :param cid: The cache id to use.
        :param idx: The cache entry index.

        :returns: True if (cid:idx) is in the cache, false otherwise.
        """
        if cid not in self.caches:
            return False
        cache = self.caches[cid]
        return idx in cache

    def get(self, cid: int, idx: int) -> QImage:
        """
        Retrieve an entry from the cache.

        :param cid: The cache id to use.
        :param idx: The cache entry index.

        :returns: The cache entry or None if it does not exist.
        """
        if cid not in self.caches:
            return None
        cache = self.caches[cid]
        if idx not in cache:
            return None
        return cache[idx]

    def add(self, cid: int, idx: int, image: QImage, buffer: bytes) -> bool:
        """
        Add an entry to the cache.
        """
        if cid not in self.caches:
            self.caches[cid] = {}
        cache = self.caches[cid]
        cache[idx] = (image, buffer)

    def evict(self, cid: int, idx: int) -> bool:
        """
        Evict an entry from the cache.

        :param cid: The cache id to use.
        :param idx: The cache entry index.

        :returns: True if an entry was evicted, false otherwise.
        """
        if not self.has(cid, idx):
            return False
        del self.caches[cid][idx]


class BrushCache:
    """Brush cache."""

    def __init__(self):
        self.entries = {}

    def has(self, idx: int) -> bool:
        return idx in self.entries

    def get(self, idx: int) -> QBrush:
        if idx in self.entries:
            return self.entries[idx]
        else:
            return None

    def add(self, idx: int, b: QBrush):
        self.entries[idx] = b


class PaletteCache:
    """ColorTable cache."""

    def __init__(self):
        self.entries = {}

    def has(self, idx: int) -> bool:
        return idx in self.entries

    def get(self, idx: int) -> List[int]:
        if idx in self.entries:
            return self.entries[idx]
        else:
            return None

    def add(self, idx: int, colors: List[int]):
        self.entries[idx] = colors


class NineGridCache:
    """NineGrid bitmap cache."""

    def __init__(self):
        self.entries = {}

    def has(self, idx: int) -> bool:
        return idx in self.entries

    def get(self, idx: int) -> QImage:
        if idx in self.entries:
            return self.entries[idx]
        else:
            return None

    def add(self, idx: int, bmp: QImage):
        self.entries[idx] = bmp


class GlyphEntry:
    """Glyph cache entry."""

    def __init__(self, glyph: Glyph):
        """Construct a cache entry from a glyph."""

        # Glyph origin.
        self.x = glyph.x
        self.y = glyph.y
        self.w = glyph.w
        self.h = glyph.h

        self.bitmap = QBitmap.fromData(QSize(self.w, self.h), glyph.data, QImage.Format_Mono)


class GlyphCache:
    """Glyph cache."""

    def __init__(self):
        self.caches = {}
        self.fragments = {}

    def get(self, cid: int, idx: int) -> GlyphEntry:
        """
        Retrieve an entry from the cache.

        :param cid: The cache id to use.
        :param idx: The cache entry index.

        :returns: The cache entry or None if it does not exist.
        """
        if cid not in self.caches:
            return None
        cache = self.caches[cid]
        if idx not in cache:
            return None
        return cache[idx]

    def add(self, cid: int, idx: int, entry: GlyphEntry) -> bool:
        """
        Add an entry to the cache.

        :returns: True if the entry is a fresh entry, False if it replaced an existing one.
        """
        if cid not in self.caches:
            self.caches[cid] = {}
        cache = self.caches[cid]
        cache[idx] = entry

    def getFragment(self, cid: int, fid: int) -> bytes:
        """Get a glyph fragment."""
        if cid not in self.fragments:
            return None
        cache = self.fragments[cid]
        if fid not in cache:
            return None
        return cache[fid]

    def addFragment(self, cid: int, fid: int, frag: bytes):
        """Store a glyph fragment."""
        if cid not in self.fragments:
            self.fragments[cid] = {}

        cache = self.fragments[cid]
        cache[fid] = frag



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Contains custom logging handlers for the library.
"""

import binascii
import json
import logging
from datetime import datetime


class VariableFormatter(logging.Formatter):
    """
    Formatter class that provides custom format variables with default values.
    """

    def __init__(self, fmt: str = None, datefmt: str = None, style: str = "%", defaultVariables: dict = None):
        super().__init__(fmt = fmt, datefmt = datefmt, style = style)
        self.defaultVariables = defaultVariables if defaultVariables is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        for variable, value in self.defaultVariables.items():
            if not hasattr(record, variable):
                setattr(record, variable, value)

        return super().format(record)


class JSONFormatter(logging.Formatter):
    """
    Formatter that returns a single JSON line of the provided data.
    Example usage: logger.info("MITM Server listening on port %(port)d", {"port": listenPort})
    Will output: {"message": "MITM Server listening on port %(port)d", "loggerName": "mitm", "timestamp": "2018-12-03T10:51:S.f-0500", "level": "INFO", "port": 3388}
    """

    def __init__(self, baseDict: dict = None):
        """
        :param baseDict: dictionary with base values that should be in every log message.
        """
        super().__init__()
        self.baseDict = baseDict if baseDict is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        data = self.baseDict.copy()

        data.update({
            "message": record.msg,
            "loggerName": record.name,
            "timestamp": datetime.strftime(datetime.utcfromtimestamp(record.created), "%Y-%m-%dT%H:%M:%S.%f"),
            "level": record.levelname,
        })

        if hasattr(record, "sessionID"):
            data.update({
                "sessionID": record.sessionID
            })

        if hasattr(record, "clientIp"):
            data.update({
                "clientIp": record.clientIp
            })

        if isinstance(record.args, dict):
            data.update(record.args)

        return json.dumps(data, ensure_ascii=False, default=lambda item: item.__repr__())


class SSLSecretFormatter(logging.Formatter):
    """
    Custom formatter used to log SSL client randoms and master secrets.
    """

    def __init__(self):
        super().__init__()

    def format(self, record: logging.LogRecord) -> str:
        return "CLIENT_RANDOM {} {}".format(binascii.hexlify(record.msg).decode(),
                                            binascii.hexlify(record.args[0]).decode())


class NTLMSSPHashFormatter(logging.Formatter):
    """
    Custom formatter used to log NTLMSSP hashes.
    """

    @staticmethod
    def formatNTLMSSPHash(user: str, domain: str, serverChallenge: bytes, proof: bytes, response: bytes) -> str:
        return f"{user}::{domain}:{serverChallenge.hex()}:{proof.hex()}:{response.hex()}"

    def format(self, record: logging.LogRecord) -> str:
        user = record.msg
        domain, serverChallenge, proof, response = record.args[0 : 4]
        return NTLMSSPHashFormatter.formatNTLMSSPHash(user, domain, serverChallenge, proof, response)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.logging.adapters import SessionLogger
from pyrdp.logging.filters import ConnectionMetadataFilter, LoggerNameFilter, SensorFilter
from pyrdp.logging.formatters import JSONFormatter, VariableFormatter
from pyrdp.logging.handlers import NotifyHandler
from pyrdp.logging.log import getSSLLogger, LOGGER_NAMES, configure
from pyrdp.logging.rc4 import RC4LoggingObserver



#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from PySide6.QtGui import QPainter

# All raster operations defined by [MS-RDPEGDI] Section 2.2.2.2.1.1.1.7
BLACKNESS = 0x00
DPSoon = 0x01
DPSona = 0x02
PSon = 0x03
SDPona = 0x04
DPon = 0x05
PDSxnon = 0x06
PDSaon = 0x07
SDPnaa = 0x08
PDSxon = 0x09
DPna = 0x0A
PSDnaon = 0x0B
SPna = 0x0C
PDSnaon = 0x0D
PDSonon = 0x0E
Pn = 0x0F
PDSona = 0x10
NOTSRCERASE = 0x11
SDPxnon = 0x12
SDPaon = 0x13
DPSxnon = 0x14
DPSaon = 0x15
PSDPSanaxx = 0x16
SSPxDSxaxn = 0x17
SPxPDxa = 0x18
SDPSanaxn = 0x19
PDSPaox = 0x1A
SDPSxaxn = 0x1B
PSDPaox = 0x1C
DSPDxaxn = 0x1D
PDSox = 0x1E
PDSoan = 0x1F
DPSnaa = 0x20
SDPxon = 0x21
DSna = 0x22
SPDnaon = 0x23
SPxDSxa = 0x24
PDSPanaxn = 0x25
SDPSaox = 0x26
SDPSxnox = 0x27
DPSxa = 0x28
PSDPSaoxxn = 0x29
DPSana = 0x2A
SSPxPDxaxn = 0x2B
SPDSoax = 0x2C
PSDnox = 0x2D
PSDPxox = 0x2E
PSDnoan = 0x2F
PSna = 0x30
SDPnaon = 0x31
SDPSoox = 0x32
NOTSRCCOPY = 0x33
SPDSaox = 0x34
SPDSxnox = 0x35
SDPox = 0x36
SDPoan = 0x37
PSDPoax = 0x38
SPDnox = 0x39
SPDSxox = 0x3A
SPDnoan = 0x3B
PSx = 0x3C
SPDSonox = 0x3D
SPDSnaox = 0x3E
PSan = 0x3F
PSDnaa = 0x40
DPSxon = 0x41
SDxPDxa = 0x42
SPDSanaxn = 0x43
SRCERASE = 0x44
DPSnaon = 0x45
DSPDaox = 0x46
PSDPxaxn = 0x47
SDPxa = 0x48
PDSPDaoxxn = 0x49
DPSDoax = 0x4A
PDSnox = 0x4B
SDPana = 0x4C
SSPxDSxoxn = 0x4D
PDSPxox = 0x4E
PDSnoan = 0x4F
PDna = 0x50
DSPnaon = 0x51
DPSDaox = 0x52
SPDSxaxn = 0x53
DPSonon = 0x54
DSTINVERT = 0x55
DPSox = 0x56
DPSoan = 0x57
PDSPoax = 0x58
DPSnox = 0x59
PATINVERT = 0x5A
DPSDonox = 0x5B
DPSDxox = 0x5C
DPSnoan = 0x5D
DPSDnaox = 0x5E
DPan = 0x5F
PDSxa = 0x60
DSPDSaoxxn = 0x61
DSPDoax = 0x62
SDPnox = 0x63
SDPSoax = 0x64
DSPnox = 0x65
SRCINVERT = 0x66
SDPSonox = 0x67
DSPDSonoxxn = 0x68
PDSxxn = 0x69
DPSax = 0x6A
PSDPSoaxxn = 0x6B
SDPax = 0x6C
PDSPDoaxxn = 0x6D
SDPSnoax = 0x6E
PDSxnan = 0x6F
PDSana = 0x70
SSDxPDxaxn = 0x71
SDPSxox = 0x72
SDPnoan = 0x73
DSPDxox = 0x74
DSPnoan = 0x75
SDPSnaox = 0x76
DSan = 0x77
PDSax = 0x78
DSPDSoaxxn = 0x79
DPSDnoax = 0x7A
SDPxnan = 0x7B
SPDSnoax = 0x7C
DPSxnan = 0x7D
SPxDSxo = 0x7E
DPSaan = 0x7F
DPSaa = 0x80
SPxDSxon = 0x81
DPSxna = 0x82
SPDSnoaxn = 0x83
SDPxna = 0x84
PDSPnoaxn = 0x85
DSPDSoaxx = 0x86
PDSaxn = 0x87
SRCAND = 0x88
SDPSnaoxn = 0x89
DSPnoa = 0x8A
DSPDxoxn = 0x8B
SDPnoa = 0x8C
SDPSxoxn = 0x8D
SSDxPDxax = 0x8E
PDSanan = 0x8F
PDSxna = 0x90
SDPSnoaxn = 0x91
DPSDPoaxx = 0x92
SPDaxn = 0x93
PSDPSoaxx = 0x94
DPSaxn = 0x95
DPSxx = 0x96
PSDPSonoxx = 0x97
SDPSonoxn = 0x98
DSxn = 0x99
DPSnax = 0x9A
SDPSoaxn = 0x9B
SPDnax = 0x9C
DSPDoaxn = 0x9D
DSPDSaoxx = 0x9E
PDSxan = 0x9F
DPa = 0xA0
PDSPnaoxn = 0xA1
DPSnoa = 0xA2
DPSDxoxn = 0xA3
PDSPonoxn = 0xA4
PDxn = 0xA5
DSPnax = 0xA6
PDSPoaxn = 0xA7
DPSoa = 0xA8
DPSoxn = 0xA9
DSTCOPY = 0xAA
DPSono = 0xAB
SPDSxax = 0xAC
DPSDaoxn = 0xAD
DSPnao = 0xAE
DPno = 0xAF
PDSnoa = 0xB0
PDSPxoxn = 0xB1
SSPxDSxox = 0xB2
SDPanan = 0xB3
PSDnax = 0xB4
DPSDoaxn = 0xB5
DPSDPaoxx = 0xB6
SDPxan = 0xB7
PSDPxax = 0xB8
DSPDaoxn = 0xB9
DPSnao = 0xBA
MERGEPAINT = 0xBB
SPDSanax = 0xBC
SDxPDxan = 0xBD
DPSxo = 0xBE
DPSano = 0xBF
MERGECOPY = 0xC0
SPDSnaoxn = 0xC1
SPDSonoxn = 0xC2
PSxn = 0xC3
SPDnoa = 0xC4
SPDSxoxn = 0xC5
SDPnax = 0xC6
PSDPoaxn = 0xC7
SDPoa = 0xC8
SPDoxn = 0xC9
DPSDxax = 0xCA
SPDSaoxn = 0xCB
SRCCOPY = 0xCC
SDPono = 0xCD
SDPnao = 0xCE
SPno = 0xCF
PSDnoa = 0xD0
PSDPxoxn = 0xD1
PDSnax = 0xD2
SPDSoaxn = 0xD3
SSPxPDxax = 0xD4
DPSanan = 0xD5
PSDPSaoxx = 0xD6
DPSxan = 0xD7
PDSPxax = 0xD8
SDPSaoxn = 0xD9
DPSDanax = 0xDA
SPxDSxan = 0xDB
SPDnao = 0xDC
SDno = 0xDD
SDPxo = 0xDE
SDPano = 0xDF
PDSoa = 0xE0
PDSoxn = 0xE1
DSPDxax = 0xE2
PSDPaoxn = 0xE3
SDPSxax = 0xE4
PDSPaoxn = 0xE5
SDPSanax = 0xE6
SPxPDxan = 0xE7
SSPxDSxax = 0xE8
DSPDSanaxxn = 0xE9
DPSao = 0xEA
DPSxno = 0xEB
SDPao = 0xEC
SDPxno = 0xED
SRCPAINT = 0xEE
SDPnoo = 0xEF
PATCOPY = 0xF0
PDSono = 0xF1
PDSnao = 0xF2
PSno = 0xF3
PSDnao = 0xF4
PDno = 0xF5
PDSxo = 0xF6
PDSano = 0xF7
PDSao = 0xF8
PDSxno = 0xF9
DPo = 0xFA
PATPAINT = 0xFB
PSo = 0xFC
PSDnoo = 0xFD
DPSoo = 0xFE
WHITENESS = 0xFF


# Mapping to supported Qt operations.
_rop2 = [
    None,
    QPainter.RasterOp_ClearDestination,  # 0
    QPainter.RasterOp_NotSourceAndNotDestination,  # DPon
    QPainter.RasterOp_NotSourceAndDestination,  # DPna
    QPainter.RasterOp_NotSource,  # Pn
    QPainter.RasterOp_SourceAndNotDestination,  # PDna
    QPainter.RasterOp_NotDestination,  # Dn
    QPainter.RasterOp_SourceXorDestination,  # DPx
    QPainter.RasterOp_NotSourceOrNotDestination,  # DPan
    QPainter.RasterOp_SourceAndDestination,  # DPa
    QPainter.RasterOp_NotSourceXorDestination,  # DPxn
    QPainter.CompositionMode_Destination,  # D
    QPainter.RasterOp_NotSourceOrDestination,  # DPno
    QPainter.CompositionMode_Source,  # P
    QPainter.RasterOp_SourceOrNotDestination,  # PDno
    QPainter.RasterOp_SourceOrDestination,  # PDo
    QPainter.RasterOp_SetDestination,  # 1
]

_rop3 = {
    BLACKNESS: QPainter.RasterOp_ClearDestination,
    DPon: QPainter.RasterOp_NotSourceAndNotDestination,
    DPna: QPainter.RasterOp_NotSourceAndDestination,
    Pn: QPainter.RasterOp_NotSource,
    NOTSRCERASE: QPainter.RasterOp_NotSourceAndNotDestination,
    DSna: QPainter.RasterOp_NotSourceAndDestination,
    NOTSRCCOPY: QPainter.RasterOp_NotSource,
    SRCERASE: QPainter.RasterOp_SourceAndNotDestination,
    PDna: QPainter.RasterOp_SourceAndNotDestination,
    DSTINVERT: QPainter.RasterOp_NotDestination,
    PATINVERT: QPainter.RasterOp_SourceXorDestination,
    DPan: QPainter.RasterOp_NotSourceOrNotDestination,
    SRCINVERT: QPainter.RasterOp_SourceXorDestination,
    DSan: QPainter.RasterOp_NotSourceOrNotDestination,
    SRCAND: QPainter.RasterOp_SourceAndDestination,
    DSxn: QPainter.RasterOp_NotSourceXorDestination,
    DPa: QPainter.RasterOp_SourceAndDestination,
    PDxn: QPainter.RasterOp_NotSourceXorDestination,
    DSTCOPY: QPainter.CompositionMode_Destination,
    DPno: QPainter.RasterOp_NotSourceOrDestination,
    MERGEPAINT: QPainter.RasterOp_NotSourceOrDestination,
    SRCCOPY: QPainter.CompositionMode_Source,
    SDno: QPainter.RasterOp_SourceOrNotDestination,
    SRCPAINT: QPainter.RasterOp_SourceOrDestination,
    PATCOPY: QPainter.CompositionMode_Source,
    PDno: QPainter.RasterOp_SourceOrNotDestination,
    DPo: QPainter.RasterOp_SourceOrDestination,
    WHITENESS: QPainter.RasterOp_SetDestination,
    PSDPxax: QPainter.RasterOp_SourceAndDestination,
}


def set_rop3(op: int, painter: QPainter):
    """
    Configure a QPainter with the required ternary raster operation.

    :parma op: The operation identifier.
    :param painter: The painter being used.
    :returns: The operation that will be processed.
    """
    if op not in _rop3:
        return None

    mode = _rop3[op]
    painter.setCompositionMode(mode)
    return mode


def set_rop2(op: int, painter: QPainter):
    """
    Configure a QPainter with the required binary raster operation.

    :parma op: The operation identifier.
    :param painter: The painter being used.
    :returns: The operation that will be processed.
    """
    if op < 0 or op >= len(_rop2):
        return None
    mode = _rop2[op]
    painter.setCompositionMode(mode)
    return mode


def rop_slow(code: str, dst, src, pal):
    """
    Slow but generic fallback implementation of raster operations.

    This is not implemented currently as Qt directly supports most of the raster operations 
    that RDP servers send.

    This function implements the RPN-notation described in [MS-RDPEGDI][1]
    with a generic stack machine. It is much slower than having a hardcoded
    and optimized function for a particular operation, but greatly reduces
    the amount of code required.

    [1]: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpegdi/a9a85075-e796-45eb-b84a-f399324a1109
    """

    stack = []
    for c in code:
        if c == 'D':
            stack.append(dst)
        elif c == 'S':
            stack.append(src)
        elif c == 'P':
            if pal is None:
                raise SyntaxError('Palette is not present.')
            stack.append(pal)
        else:
            lhs = stack.pop()
            rhs = None
            res = lhs  # TODO: Actually perform the operation.

            if c != 'n':
                rhs = stack.pop()

            if c == 'x':  # XOR
                print(f'{lhs} ^ {rhs}')
            elif c == 'n':  # NOT
                print(f'~{lhs}')
            elif c == 'a':  # AND
                print(f'{lhs} & {rhs}')
            elif c == 'o':  # OR
                print(f'{lhs} | {rhs}')

            stack.append(res)
    out = stack.pop()

    assert len(stack) == 0
    return out



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from PySide6.QtWidgets import QPushButton, QWidget
from PySide6.QtGui import QIcon


class PlayPauseButton(QPushButton):
    """
    Button that switches between "play" and "pause" depending on its state.
    """

    def __init__(self, parent: QWidget = None, playText = "Play", pauseText = "Pause"):
        super().__init__(parent)
        self.clicked.connect(self.onClick)
        self.playText = playText
        self.pauseText = pauseText
        self.playing = False
        self.setPlaying(self.playing)
        self.setFixedWidth(100)

    def onClick(self):
        self.setPlaying(not self.playing)

    def setPlaying(self, playing):
        self.playing = playing

        if self.playing:
            self.setText(self.pauseText)
            self.setIcon(QIcon.fromTheme("media-playback-pause"))
        else:
            self.setText(self.playText)
            self.setIcon(QIcon.fromTheme("media-playback-start"))



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.core.FileProxy import FileProxy, FileProxyObserver
from pyrdp.core.defer import defer
from pyrdp.core.event import EventEngine
from pyrdp.core.helpers import decodeUTF16LE, encodeUTF16LE, FilePositionGuard
from pyrdp.core.observer import CompositeObserver, Observer
from pyrdp.core.packing import Int16BE, Int16LE, Int32BE, Int32LE, Int8, Uint16BE, Uint16LE, Uint32BE, Uint32LE, \
    Uint64LE, Uint8
from pyrdp.core.sequencer import AsyncIOSequencer, Sequencer
from pyrdp.core.stream import ByteStream, StrictStream
from pyrdp.core.subject import ObservedBy, Subject
from pyrdp.core.timer import Timer
from pyrdp.core.twisted import AwaitableClientFactory, connectTransparent



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import TextIOBase
from sys import stdout

from pyrdp.player import BaseEventHandler


class HeadlessEventHandler(BaseEventHandler):
    """
    Handle events from a replay file without rendering to a surface.

    This event handler does not require any graphical dependencies.
    """

    def __init__(self, output: TextIOBase = stdout):
        super().__init__()
        self.output = output

    def writeText(self, text: str):
        self.output.write(text.rstrip("\x00"))

    def writeSeparator(self):
        self.output.write("\n--------------------\n")

    def onMouseButton(self, buttons, pos):
        pressed = []
        if 1 in buttons and buttons[1]:
            pressed.append('Left')
        if 2 in buttons and buttons[2]:
            pressed.append('Right')
        if 3 in buttons and buttons[3]:
            pressed.append('Middle')

        (x, y) = pos
        self.writeText(f'\n<Click ({", ".join(pressed)}) @ ({x}, {y})>')



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import decodeUTF16LE, encodeUTF16LE, StrictStream, Uint16LE, Uint32LE
from pyrdp.enum import ClientInfoFlags
from pyrdp.parser.parser import Parser
from pyrdp.pdu import ClientExtraInfo, ClientInfoPDU


class ClientInfoParser(Parser):
    """
    Read and write the RDP ClientInfo PDU which contains very useful information.
    See https://msdn.microsoft.com/en-us/library/cc240475.aspx
    """

    def doParse(self, data: bytes) -> ClientInfoPDU:
        """
        Decode a Client Info PDU from bytes.
        :param data: the Client Info PDU bytes.
        """
        stream = BytesIO(data)
        codePage = Uint32LE.unpack(stream)
        flags = Uint32LE.unpack(stream)

        isUnicode = flags & ClientInfoFlags.INFO_UNICODE != 0
        hasNullBytes = codePage == 1252 or isUnicode
        nullByteCount = 1 if hasNullBytes else 0
        unicodeMultiplier = 2 if isUnicode else 0

        domainLength = Uint16LE.unpack(stream) + nullByteCount * unicodeMultiplier
        usernameLength = Uint16LE.unpack(stream) + nullByteCount * unicodeMultiplier
        passwordLength = Uint16LE.unpack(stream) + nullByteCount * unicodeMultiplier
        alternateShellLength = Uint16LE.unpack(stream) + nullByteCount * unicodeMultiplier
        workingDirLength = Uint16LE.unpack(stream) + nullByteCount * unicodeMultiplier

        domain = decodeUTF16LE(stream.read(domainLength))
        username = decodeUTF16LE(stream.read(usernameLength))
        password = decodeUTF16LE(stream.read(passwordLength))
        alternateShell = decodeUTF16LE(stream.read(alternateShellLength))
        workingDir = decodeUTF16LE(stream.read(workingDirLength))

        extraInfoBytes = stream.read()

        if extraInfoBytes != b"":
            extraInfo = self.parseExtraInfo(extraInfoBytes)
        else:
            extraInfo = None

        return ClientInfoPDU(codePage, flags, domain, username, password, alternateShell, workingDir, extraInfo)

    def write(self, pdu: ClientInfoPDU) -> bytes:
        """
        Encode a Client Info PDU to bytes.
        :param pdu: the Client Info PDU.
        """
        stream = BytesIO()
        stream.write(Uint32LE.pack(pdu.codePage))
        stream.write(Uint32LE.pack(pdu.flags))

        isUnicode = pdu.flags & ClientInfoFlags.INFO_UNICODE != 0
        hasNullBytes = pdu.codePage == 1252 or isUnicode
        nullByteCount = 1 if hasNullBytes else 0
        unicodeMultiplier = 2 if isUnicode else 0

        domain = pdu.domain + "\x00" * nullByteCount
        username = pdu.username + "\x00" * nullByteCount
        password = pdu.password + "\x00" * nullByteCount
        alternateShell = pdu.alternateShell + "\x00" * nullByteCount
        workingDir = pdu.workingDir + "\x00" * nullByteCount

        if isUnicode:
            domain = encodeUTF16LE(domain)
            username = encodeUTF16LE(username)
            password = encodeUTF16LE(password)
            alternateShell = encodeUTF16LE(alternateShell)
            workingDir = encodeUTF16LE(workingDir)
        else:
            domain = domain.encode()
            username = username.encode()
            password = password.encode()
            alternateShell = alternateShell.encode()
            workingDir = workingDir.encode()

        domainLength = len(domain) - nullByteCount * unicodeMultiplier
        usernameLength = len(username) - nullByteCount * unicodeMultiplier
        passwordLength = len(password) - nullByteCount * unicodeMultiplier
        alternateShellLength = len(alternateShell) - nullByteCount * unicodeMultiplier
        workingDirLength = len(workingDir) - nullByteCount * unicodeMultiplier

        stream.write(Uint16LE.pack(domainLength))
        stream.write(Uint16LE.pack(usernameLength))
        stream.write(Uint16LE.pack(passwordLength))
        stream.write(Uint16LE.pack(alternateShellLength))
        stream.write(Uint16LE.pack(workingDirLength))
        stream.write(domain)
        stream.write(username)
        stream.write(password)
        stream.write(alternateShell)
        stream.write(workingDir)

        if pdu.extraInfo is not None:
            extraInfoBytes = self.writeExtraInfo(pdu.extraInfo)
            stream.write(extraInfoBytes)

        return stream.getvalue()

    def parseExtraInfo(self, data: bytes) -> ClientExtraInfo:
        stream = BytesIO(data)
        clientAddressFamily = Uint16LE.unpack(stream)

        clientAddressLength = Uint16LE.unpack(stream)
        clientAddress = stream.read(clientAddressLength)

        clientDirLength = Uint16LE.unpack(stream)
        clientDir = stream.read(clientDirLength)

        extraInfo = ClientExtraInfo(clientAddressFamily, clientAddress, clientDir)
        stream = StrictStream(stream)

        try:
            extraInfo.clientTimeZone = stream.read(172)
            extraInfo.clientSessionID = Uint32LE.unpack(stream)
            extraInfo.performanceFlags = Uint32LE.unpack(stream)

            autoReconnectCookieLength = Uint16LE.unpack(stream)
            extraInfo.autoReconnectCookie = stream.read(autoReconnectCookieLength)

            stream.read(4)

            dynamicDSTTimeZoneKeyNameLength = Uint16LE.unpack(stream)
            extraInfo.dynamicDSTTimeZoneKeyName = stream.read(dynamicDSTTimeZoneKeyNameLength)

            extraInfo.dynamicDaylightTimeDisabled = bool(Uint16LE.unpack(stream))
        except EOFError:
            pass

        return extraInfo

    def writeExtraInfo(self, extraInfo: ClientExtraInfo) -> bytes:
        stream = BytesIO()
        Uint16LE.pack(extraInfo.clientAddressFamily, stream)

        Uint16LE.pack(len(extraInfo.clientAddress), stream)
        stream.write(extraInfo.clientAddress)

        Uint16LE.pack(len(extraInfo.clientDir), stream)
        stream.write(extraInfo.clientDir)

        if extraInfo.clientTimeZone is None:
            return stream.getvalue()

        stream.write(extraInfo.clientTimeZone)

        if extraInfo.clientSessionID is None:
            return stream.getvalue()

        Uint32LE.pack(extraInfo.clientSessionID, stream)

        if extraInfo.performanceFlags is None:
            return stream.getvalue()

        Uint32LE.pack(extraInfo.performanceFlags, stream)

        if extraInfo.autoReconnectCookie is None:
            return stream.getvalue()

        Uint16LE.pack(len(extraInfo.autoReconnectCookie), stream)
        stream.write(extraInfo.autoReconnectCookie)

        stream.write(b"\x00" * 4)

        if extraInfo.dynamicDSTTimeZoneKeyName is None:
            return stream.getvalue()

        Uint16LE.pack(len(extraInfo.dynamicDSTTimeZoneKeyName), stream)
        stream.write(extraInfo.dynamicDSTTimeZoneKeyName)

        if extraInfo.dynamicDaylightTimeDisabled is None:
            return stream.getvalue()

        Uint16LE.pack(int(extraInfo.dynamicDaylightTimeDisabled), stream)
        return stream.getvalue()


#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
import logging
from binascii import hexlify

from twisted.internet.protocol import connectionDone, Protocol

from pyrdp.core import ObservedBy
from pyrdp.exceptions import ParsingError, ExploitError
from pyrdp.layer.layer import IntermediateLayer, LayerObserver
from pyrdp.logging import LOGGER_NAMES, getSSLLogger
from pyrdp.parser.tcp import TCPParser
from pyrdp.pdu import PDU


class TCPObserver(LayerObserver):
    def onConnection(self):
        """
        Called when a TCP connection is made.
        """
        pass

    def onDisconnection(self, reason):
        """
        Called when the TCP connection is lost.
        :param reason: reason for disconnection.
        """
        pass


@ObservedBy(TCPObserver)
class TwistedTCPLayer(IntermediateLayer, Protocol):
    """
    Twisted protocol class and first layer in a stack.
    ObservedBy: TCPObserver
    Never notifies observers about PDUs because there isn't really a TCP PDU type per se.
    TCP observers are notified when a connection is made.
    """

    def __init__(self):
        self.log = logging.getLogger(LOGGER_NAMES.PYRDP)
        super().__init__(TCPParser())
        self.connectedEvent = asyncio.Event()
        self.logSSLRequired = False

    def logSSLParameters(self):
        """
        Log the SSL parameters of the connection in a format suitable for decryption by Wireshark.
        """
        getSSLLogger().info(self.transport.protocol._tlsConnection.client_random(), self.transport.protocol._tlsConnection.master_key())

    def connectionMade(self):
        """
        When the TCP handshake is completed, notify the observer.
        """
        self.connectedEvent.set()
        self.observer.onConnection()

    def connectionLost(self, reason=connectionDone):
        """
        :param reason: reason for disconnection.
        """
        self.observer.onDisconnection(reason)

    def disconnect(self, abort = False):
        """
        Close the TCP connection.
        :param abort: True to force close the connection, False to end gracefully.
        """

        if self.transport:
            if abort:
                self.transport.abortConnection()
            else:
                self.transport.loseConnection()

    def dataReceived(self, data: bytes):
        """
        Called whenever data is received.
        :param data: bytes received.
        """
        try:
            if self.logSSLRequired:
                self.logSSLParameters()
                self.logSSLRequired = False

            self.recv(data)
        except KeyboardInterrupt:
            raise
        except ExploitError as e:
            # Ideally it would be nice to have a system for detecting exploits without interrupting the connection
            self.log.info("Exploit detected: %(exploitInfo)s. %(parserInfo)s", {
                "exploitInfo": str(e),
                "parserInfo": e.formatLayer(len(e.layers) - 1)
            })
        except Exception as e:
            self.log.exception(e)

            if isinstance(e, ParsingError):
                self.log.error("Parser information: %(parserInfo)s", {"parserInfo": e.formatLayer(len(e.layers) - 1)})

            self.log.error("Exception occurred when receiving: %(exceptionData)s" , {"exceptionData": hexlify(data).decode()})

            raise

    def sendBytes(self, data: bytes):
        """
        Send raw TCP bytes.
        :param data: bytes to send.
        """
        self.transport.write(data)

    def startTLS(self, tlsContext):
        """
        Perform a TLS handshake so that all further communications are encrypted.
        :param tlsContext: Twisted TLS Context object (like DefaultOpenSSLContextFactory)
        """
        self.logSSLRequired = True
        self.transport.startTLS(tlsContext)

    def shouldForward(self, pdu: PDU) -> bool:
        return True


@ObservedBy(TCPObserver)
class AsyncIOTCPLayer(IntermediateLayer, asyncio.Protocol):
    """
    AsyncIO protocol class and first layer in a stack.
    ObservedBy: TCPObserver
    Never notifies observers about PDUs because there isn't really a TCP PDU type per se.
    TCP observers are notified when a connection is made.
    """

    def __init__(self):
        super().__init__(TCPParser())
        self.connectedEvent = asyncio.Event()
        self.transport: asyncio.Transport = None

    def connection_made(self, transport: asyncio.BaseTransport):
        """
        When the TCP handshake is completed, notify the observer.
        """
        self.transport = transport
        self.connectedEvent.set()
        self.observer.onConnection()

    def connection_lost(self, exception=connectionDone):
        """
        :param exception: reason for disconnection.
        """
        self.observer.onDisconnection(exception)

    def disconnect(self, abort = False):
        """
        Close the TCP connection.
        :param abort: True to force close the connection, False to end gracefully.
        """
        if abort:
            self.transport.abort()
        else:
            self.transport.close()

    def data_received(self, data: bytes):
        """
        Called whenever data is received.
        :param data: bytes received.
        """

        try:
            self.recv(data)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            logging.getLogger(LOGGER_NAMES.PYRDP).exception(e)
            raise

    def sendBytes(self, data: bytes):
        """
        Send raw TCP bytes.
        :param data: bytes to send.
        """
        self.transport.write(data)

    def shouldForward(self, pdu: PDU) -> bool:
        return True



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.pdu import PDU


class Observer:
    """
    Base observer class used across PyRDP.
    """
    def __init__(self, **kwargs):
        """
        Initialize a new Observer object.
        The kwargs are used to allow users to define custom handlers and pass them as arguments instead of inheriting from an observer class.
        This is useful in case of multiple inheritance, because only one method of the same name is preserved.
        """
        self.peer: Observer = None
        for (name, handler) in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, handler)
            else:
                raise TypeError("Unexpected keyword argument '%s'" % name)

    def setPeer(self, peer: 'Observer'):
        self.peer = peer
        peer.peer = self

    def onPDUReceived(self, pdu: PDU):
        pass


class CompositeObserver:
    """
    Observer class that contains other observers and delegates method calls to them.
    """
    def __init__(self):
        self.observers = []

    def __getattr__(self, item):
        """
        Creates a CompositeObserverCall object, which will invoke doCall when it is called.
        """
        return CompositeObserverCall(self, item)

    def __nonzero__(self):
        return True

    def doCall(self, item, args, kwargs):
        """
        When a method is called, invoke the same method on every observer object.
        """
        for observer in self.observers:
            getattr(observer, item)(*args, **kwargs)

    def addObserver(self, observer: Observer):
        """
        Add an observer to the composite.
        """
        self.observers.append(observer)

    def removeObserver(self, observer: Observer):
        """
        Remove an observer from the composite.
        """
        self.observers.remove(observer)

class CompositeObserverCall:
    """
    Object that calls back to the CompositeObserver when it is called.
    """
    def __init__(self, composite, item):
        self.composite = composite
        self.item = item

    def __call__(self, *args, **kwargs):
        """
        Delegate the call to the composite.
        """
        self.composite.doCall(self.item, args, kwargs)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum, Enum


class ClipboardMessageType(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241097.aspx
    """

    CB_MONITOR_READY = 0x0001
    CB_FORMAT_LIST = 0x0002
    CB_FORMAT_LIST_RESPONSE = 0x0003
    CB_FORMAT_DATA_REQUEST = 0x0004
    CB_FORMAT_DATA_RESPONSE = 0x0005
    CB_TEMP_DIRECTORY = 0x0006
    CB_CLIP_CAPS = 0x0007
    CB_FILECONTENTS_REQUEST = 0x0008
    CB_FILECONTENTS_RESPONSE = 0x0009
    CB_LOCK_CLIPDATA = 0x000A
    CB_UNLOCK_CLIPDATA = 0x000B


class ClipboardMessageFlags(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241097.aspx
    """
    NONE = 0
    CB_RESPONSE_OK = 0x0001
    CB_RESPONSE_FAIL = 0x0002
    CB_ASCII_NAMES = 0x0004


class ClipboardFormatNumber(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241079.aspx
    """
    GENERIC = 13
    PALETTE = 9
    METAFILE = 3


class FileDescriptorFlags(Enum):
    """
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpeclip/a765d784-2b39-4b88-9faa-88f8666f9c35
    """
    FD_ATTRIBUTES = 0x04
    FD_FILESIZE = 0x40
    FD_WRITESTIME = 0x20
    FD_SHOWPROGRESSUI = 0x4000


class FileContentsFlags(IntEnum):
    SIZE = 0x1
    RANGE = 0x2


class ClipboardFormatName(Enum):
    """
    https://msdn.microsoft.com/en-us/library/cc241079.aspx
    """
    FILE_LIST = "FileGroupDescriptorW"
    DROP_EFFECT = "Preferred DropEffect"
    FILE_CONTENT = "FileContents"



#
# Copyright (c) 2014-2020 Sylvain Peyrefitte
# Copyright (c) 2020-2021 GoSecure Inc.
#
# This file is part of the PyRDP project.
#
# Licensed under the GPLv3 or later.
#

from typing import Tuple
from os import path

import OpenSSL
from OpenSSL import SSL

from twisted.internet import ssl


class ClientTLSContext(ssl.ClientContextFactory):
    """
    @summary: client context factory for open ssl
    """

    def getContext(self):
        # Allow the MITM to connect to an RDP Server with ANY TLS version supported by the installed
        # OpenSSL version. See https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=784153
        # It was removed from OpenSSL, but PyOpenSSL has not changed their constant names yet.
        context = SSL.Context(SSL.SSLv23_METHOD)
        context.set_options(SSL.OP_DONT_INSERT_EMPTY_FRAGMENTS)
        context.set_options(SSL.OP_TLS_BLOCK_PADDING_BUG)

        # We disable TLS 1.3 because the way to decrypt TLS 1.3 traffic differs from
        # previous TLS versions and is not yet supported by PyRDP.
        context.set_options(SSL.OP_NO_TLSv1_3)
        return context


class ServerTLSContext(ssl.DefaultOpenSSLContextFactory):
    """
    @summary: Server context factory for open ssl
    @param privateKeyFileName: Name of a file containing a private key
    @param certificateFileName: Name of a file containing a certificate
    """

    def __init__(self, privateKeyFileName, certificateFileName):
        class TPDUSSLContext(SSL.Context):
            def __init__(self, method):
                SSL.Context.__init__(self, method)
                self.set_options(SSL.OP_DONT_INSERT_EMPTY_FRAGMENTS)
                self.set_options(SSL.OP_TLS_BLOCK_PADDING_BUG)

                # See comment in ClientTLSContext
                self.set_options(SSL.OP_NO_TLSv1_3)

        # See comment in ClientTLSContext
        ssl.DefaultOpenSSLContextFactory.__init__(self, privateKeyFileName, certificateFileName, SSL.SSLv23_METHOD,
                                                  TPDUSSLContext)


class CertificateCache():
    """
    Handle multiple certificates.
    """

    def __init__(self, cachedir, log):
        self._root = cachedir
        self.log = log

    def clone(self, cert: OpenSSL.crypto.X509) -> Tuple[OpenSSL.crypto.PKey, OpenSSL.crypto.X509]:
        """Clone the provided certificate."""

        # Generate a private key for the server.
        key = OpenSSL.crypto.PKey()
        key.generate_key(OpenSSL.crypto.TYPE_RSA, cert.get_pubkey().bits())

        # Actual type is str, but this prevents warnings
        digestAlgorithm: bytes = cert.get_signature_algorithm().decode()

        # Force digest algorithm to be sha256
        if digestAlgorithm in ["md4", "md5"]:
            digestAlgorithm = "sha256"

        cert.set_pubkey(key)
        cert.sign(key, digestAlgorithm)

        return key, cert

    def lookup(self, cert: OpenSSL.crypto.X509) -> Tuple[str, str]:
        subject = cert.get_subject()
        parts = dict(subject.get_components())
        commonName = parts[b'CN'].decode()
        base = str(self._root / commonName)

        if path.exists(base + '.pem'):
            self.log.info('Using cached certificate for %(commonName)s', {'commonName': commonName})

            # Recover cache entry from disk.
            privKey = base + '.pem'
            certFile = base + '.crt'
            return privKey, certFile
        else:
            priv, cert = self.clone(cert)
            privKey = base + '.pem'
            certFile = base + '.crt'

            # Save Certificate to disk
            with open(certFile, "wb") as f:
                f.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))

            with open(privKey, "wb") as f:
                f.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, priv))

            self.log.info('Cloned server certificate to %(certFile)s', {'certFile': certFile})

            return privKey, certFile



#!/usr/bin/python3

#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import argparse
import logging
import sys
from pathlib import Path

from pyrdp.convert.PCAPConverter import PCAPConverter
from pyrdp.convert.ReplayConverter import ReplayConverter
from pyrdp.convert.utils import HANDLERS, loadSecrets
from pyrdp.player import HAS_GUI


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Path to a .pcap or .pyrdp file. "
                                      "Converting from a .pcap will always extract file transfer artifacts in addition to the actual replay.")
    parser.add_argument(
        "-l",
        "--list-only",
        help="Print the list of sessions in the capture without processing anything",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--secrets",
        help="Path to the file containing the SSL secrets to decrypt Transport Layer Security.",
    )
    parser.add_argument(
        "-f",
        "--format",
        help="Format of the output",
        choices=HANDLERS.keys(),
        default="replay",
    )
    parser.add_argument(
        "--src",
        help="If specified, limits the converted streams to connections initiated from this address",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--dst",
        help="If specified, limits the converted streams to connections destined to this address",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Path to write the converted files to. If a file name is specified, it will be used as a prefix, "
        "otherwise the result is output next to the source file with the proper extension. "
        "However if the source of the conversion is a .pcap then this option will create a directory where all files will be stored.",
    )

    args = parser.parse_args()

    if not HAS_GUI and args.format == "mp4":
        sys.stderr.write("Error: MP4 conversion requires the full PyRDP installation.")
        sys.exit(1)
    elif HAS_GUI and args.format == "mp4":
        # Initialize QT because QBitmap will segfault without it (#378, #428)
        from PySide6.QtWidgets import QApplication
        QApplication()

    logging.basicConfig(level=logging.CRITICAL)
    logging.getLogger("scapy").setLevel(logging.ERROR)

    inputFile = Path(args.input)

    if args.output:
        output = Path(args.output)
        # Pcaps create directory structures since they also extract transfered artifacts
        if output.is_dir() or inputFile.suffix in [".pcap"]:
            outputPrefix = str(output.absolute()) + "/"
        else:
            outputPrefix = str(output.parent.absolute() / output.stem) + "-"
    else:
        outputPrefix = ""

    if inputFile.suffix in [".pcap"]:
        secrets = loadSecrets(args.secrets) if args.secrets else None
        converter = PCAPConverter(inputFile, outputPrefix, args.format, secrets=secrets, srcFilter=args.src, dstFilter=args.dst, listOnly=args.list_only)
    elif inputFile.suffix in [".pyrdp"]:
        if args.format == "replay":
            sys.stderr.write("Refusing to convert a replay file to a replay file. Choose another format.")
            sys.exit(1)

        converter = ReplayConverter(inputFile, outputPrefix, args.format)
    else:
        sys.stderr.write("Unknown file extension. (Supported: .pcap, .pyrdp)")
        sys.exit(1)

    exitCode = converter.process()
    sys.exit(exitCode)

if __name__ == "__main__":
    main()



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import List, Optional

from pyrdp.enum import PlayerPDUType
from pyrdp.layer import LayerChainItem
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu import PDU, InputPDU
from pyrdp.recording import Recorder


class MITMRecorder(Recorder):
    """
    Recorder subclass that avoids recording input events when input forwarding is disabled.
    """

    def __init__(self, transports: List[LayerChainItem], state: RDPMITMState):
        super().__init__(transports)
        self.state = state

    def record(self, pdu: Optional[PDU], messageType: PlayerPDUType, forceRecording: bool = False):
        """
        Record a PDU. The forceRecording param is mostly useful for recording forged PDUs (e.g: input coming from the attacker).
        :param pdu: the PDU to record.
        :param messageType: the message type.
        :param forceRecording: when set to True, records the PDU even if forwarding is disabled. Defaults to False.
        """

        if self.state.forwardInput or forceRecording or (messageType != PlayerPDUType.FAST_PATH_INPUT and not isinstance(pdu, InputPDU)):
            super().record(pdu, messageType)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class X224PDUType(IntEnum):
    """
    X224 header codes.
    """

    X224_TPDU_CONNECTION_REQUEST = 0x0E
    X224_TPDU_CONNECTION_CONFIRM = 0x0D
    X224_TPDU_DISCONNECT_REQUEST = 0x08
    X224_TPDU_DATA = 0x0F
    X224_TPDU_ERROR = 0x07



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from io import BytesIO

from pyrdp.core import FilePositionGuard
from pyrdp.exceptions import ParsingError
from pyrdp.pdu import PDU


class BaseParser:
    def parse(self, data):
        """
        Decode a PDU from data.
        :param data: PDU data.
        :return: an instance of a PDU class.
        """

        try:
            return self.doParse(data)
        except ParsingError as e:
            self.handleParsingError(e, data)
            raise

    def doParse(self, data) -> PDU:
        raise NotImplementedError("Parse is not implemented")

    def handleParsingError(self, e: ParsingError, data):
        """
        Add self and data to the list of layers of the parsing error.
        """
        raise NotImplementedError("handleParsingError is not implemented")

    def write(self, pdu: PDU) -> bytes:
        """
        Encode a PDU to bytes.
        :param pdu: instance of a PDU class.
        """
        raise NotImplementedError("Write is not implemented")


class Parser(BaseParser):
    # For type hints
    def parse(self, data: bytes):
        return super().parse(data)

    def doParse(self, data: bytes) -> PDU:
        return super().doParse(data)

    def handleParsingError(self, e: ParsingError, data: bytes):
        e.addLayer(self, data)


class StreamParser(BaseParser):
    # For type hints
    def parse(self, stream: BytesIO):
        return super().parse(stream)

    def doParse(self, stream: BytesIO) -> PDU:
        return super().doParse(stream)

    def handleParsingError(self, e: ParsingError, stream: BytesIO):
        with FilePositionGuard(stream):
            e.addLayer(self, stream.read())



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging
import logging.config
from configparser import ConfigParser

from pathlib import Path
from pyrdp.logging.filters import LoggerNameFilter


class LOGGER_NAMES:
    # Root logger
    PYRDP = "pyrdp"
    MITM = f"{PYRDP}.mitm"
    MITM_CONNECTIONS = f"{MITM}.connections"
    PLAYER = f"{PYRDP}.player"
    PLAYER_UI = f"{PLAYER}.ui"
    NTLMSSP = f"ntlmssp"

    # Independent logger
    CRAWLER = "crawler"


def getSSLLogger():
    """
    Get the SSL logger.
    """
    return logging.getLogger("ssl")


def info(*args):
    logging.getLogger(LOGGER_NAMES.PYRDP).info(*args)


def debug(*args):
    logging.getLogger(LOGGER_NAMES.PYRDP).debug(*args)


def warning(*args):
    logging.getLogger(LOGGER_NAMES.PYRDP).warning(*args)


def error(*args):
    logging.getLogger(LOGGER_NAMES.PYRDP).error(*args)


def convertConfig(cfg: ConfigParser):
    """
    Transform a config file into a dictionary.

    Keys with the title format `section:subsection:subsection` are
    transformed into nested dictionaries

    # Examples

    ```ini
    [toplevel]
    a = 1
    b = 2

    [toplevel:sublevel]
    c = 3
    d = 4
    ```

    would convert to

    ```python
    'toplevel': {
        'a': 1,
        'b': 2,
        'sublevel': {
            'c': 3,
            'd': 4,
        }
    }
    ````
    """
    def get(d: dict, key: str, create=False) -> dict:
        """Resolve a nested key in a dictionary."""
        nested = key.split(':')
        sub = None
        for n in nested:
            if sub is None:  # Top Level
                if n not in d:
                    if not create:
                        return None
                    d[n] = {}
                sub = d[n]
            else:
                if n not in sub:
                    if not create:
                        return None
                    sub[n] = {}
                sub = sub[n]
        return sub

    out = {}
    for s in cfg.sections():
        sec = get(out, s, create=True)
        for (k, v) in cfg.items(s):
            sec[k] = v
    return out


def configure(config: ConfigParser) -> bool:
    """Configure logging based on settings from disk."""
    try:
        cfg = convertConfig(config)
        logs = cfg['logs']
        logs['version'] = int(logs['version'])  # Needs to be integer.

        for (k, v) in logs['loggers'].items():
            v['handlers'] = [x.strip() for x in v['handlers'].split(',')]

        # Ensure log directory exists.
        logDir = Path(cfg['vars']['output_dir']) / cfg['vars']['log_dir']
        logDir.mkdir(exist_ok = True)

        logging.config.dictConfig(cfg['logs'])

        # Enable the user configured filter.
        if 'filter' in logs:
            root = logging.getLogger(LOGGER_NAMES.PYRDP)
            for h in root.handlers:
                # Use type() because we want specific type, not a subclass.
                if type(h) == logging.StreamHandler:
                    h.filters.clear()
                    h.addFilter(LoggerNameFilter(logs['filter']))
        return True
    except Exception as e:
        logging.warning('Error Parsing PyRDP Configuraton - %s', e)
        return False



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import LoggerAdapter

from pyrdp.enum import ErrorInfo, MCSPDUType, X224PDUType
from pyrdp.layer import FastPathObserver, LayerObserver, MCSObserver, SecurityObserver, SlowPathObserver, X224Observer
from pyrdp.parser import ClientInfoParser
from pyrdp.pdu import FastPathPDU, MCSAttachUserConfirmPDU, MCSChannelJoinConfirmPDU, MCSConnectResponsePDU, MCSPDU, \
    SecurityExchangePDU, SlowPathPDU, X224PDU


class LoggingObserver:
    """
    Base class for logging observers. Logs the string representation of PDUs.
    """
    def __init__(self, log: LoggerAdapter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = log

    def logPDU(self, pdu):
        self.log.debug("Received %(pdu)s", {"pdu": pdu})


class X224Logger(LoggingObserver, X224Observer):
    """
    Logging observer for X224 layers.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)
        self.log = log

    def onPDUReceived(self, pdu: X224PDU):
        if pdu.header in [X224PDUType.X224_TPDU_DATA]:
            self.log.debug("Received %(type)s", {"type": pdu.header})
        else:
            self.logPDU(pdu)

        super().onPDUReceived(pdu)


class MCSLogger(LoggingObserver, MCSObserver):
    """
    Logging observer for MCS layers.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)
        self.log = log

    def onPDUReceived(self, pdu: MCSPDU):
        if pdu.header in [MCSPDUType.SEND_DATA_REQUEST, MCSPDUType.SEND_DATA_INDICATION]:
            self.log.debug("Received %(type)s", {"type": pdu.header})
        else:
            self.logPDU(pdu)

        super().onPDUReceived(pdu)

    def onConnectResponse(self, pdu: MCSConnectResponsePDU):
        if pdu.result == 0:
            self.log.debug("MCS Connection successful")
        else:
            self.log.debug("MCS Connection failed")

    def onAttachUserConfirm(self, pdu: MCSAttachUserConfirmPDU):
        if pdu.result == 0:
            self.log.debug("Attach User successful")
        else:
            self.log.debug("Attach User failed")

    def onChannelJoinConfirm(self, pdu: MCSChannelJoinConfirmPDU):
        if pdu.result == 0:
            self.log.debug("Channel Join #%(channelID)s successful", {"channelID": pdu.channelID})
        else:
            self.log.debug("Channel Join #%(channelID)s failed", {"channelID": pdu.channelID})


class SecurityLogger(LoggingObserver, SecurityObserver):
    """
    Logging observer for security layers.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)
        self.log = log

    def onSecurityExchangeReceived(self, pdu: SecurityExchangePDU):
        self.logPDU(pdu)

    def onClientInfoReceived(self, data: bytes):
        pdu = ClientInfoParser().parse(data)
        self.logPDU(pdu)

    def onLicensingDataReceived(self, data: bytes):
        self.log.debug("Received licensing data")


class SlowPathLogger(LoggingObserver, SlowPathObserver):
    """
    Logging observer for slow-path layers.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)

    def onPDUReceived(self, pdu: SlowPathPDU):
        self.logPDU(pdu)

    def logPDU(self, pdu):
        if hasattr(pdu.header, "subtype"):
            if hasattr(pdu, "errorInfo"):
                description = pdu.errorInfo

                if pdu.errorInfo != ErrorInfo.ERRINFO_NONE:
                    errorInfoText = ErrorInfo.getText(pdu.errorInfo)

                    if pdu.errorInfo in [ErrorInfo.ERRINFO_LOGOFF_BY_USER]:
                        self.log.info("%(description)s", {"description": errorInfoText})
                    else:
                        self.log.error("RDP Error Info: %(description)s", {"description": errorInfoText})
            else:
                description = pdu.header.subtype
        else:
            description = pdu.header.pduType

        self.log.debug("Received %(description)s", {"description": description})


class FastPathLogger(LoggingObserver, FastPathObserver):
    """
    Logging observer for fast-path layers.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)

    def onPDUReceived(self, pdu: FastPathPDU):
        self.logPDU(pdu)


class LayerLogger(LoggingObserver, LayerObserver):
    """
    Generic logging observer for all layer types.
    """
    def __init__(self, log: LoggerAdapter):
        super().__init__(log)

    def onPDUReceived(self, pdu):
        self.logPDU(pdu)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.security.crypto import RC4Crypter
from pyrdp.security.ntlmssp import NTLMSSPState
from pyrdp.security.rc4proxy import RC4CrypterProxy
from pyrdp.security.settings import SecuritySettings, SecuritySettingsObserver



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.core.observer import CompositeObserver, Observer


class Subject:
    """
    Base class for objects that can have observers.
    """
    def __init__(self):
        super().__init__()
        self.observer = CompositeObserver()

    def addObserver(self, observer: Observer):
        """
        Add an observer to this subject.
        """
        return self.observer.addObserver(observer)

    def removeObserver(self, observer: Observer):
        """
        Remove an observer from this subject.
        """
        return self.observer.removeObserver(observer)

def ObservedBy(ObserverClass: type):
    """
    This decorator adds a `createObserver` method to a class that creates an
    observer object and forwards all keyword arguments to its constructor.

    :param ObserverClass: The observer class.
    """
    def setCreateObserverMethod(SubjectClass: type) -> type:
        """
        Add a createObserver method to a class.
        :param SubjectClass: the subject class.
        """

        def createObserver(self, **kwargs):
            """
            Creates a new observer by forwarding all keyword arguments to it.
            """
            observer = ObserverClass(**kwargs)
            self.addObserver(observer)
            return observer

        SubjectClass.createObserver = createObserver
        return SubjectClass
    
    return setCreateObserverMethod


#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
from abc import ABCMeta, abstractmethod
from typing import Callable, List, Optional

from pyrdp.core.defer import defer


class Sequencer(metaclass = ABCMeta):
    """
    Class used for spreading function calls across time.
    """

    def __init__(self, functions: List[Callable[[], Optional[int]]]):
        """
        :param functions: list of functions to be called, each one optionally returning an amount of time to wait for.
        """
        self.functions = functions

    def run(self):
        """
        Run all remaining functions.
        """

        while len(self.functions) > 0:
            wait = self.functions.pop(0)()

            if wait is not None and wait > 0:
                self.wait(wait)
                break

    @abstractmethod
    def wait(self, waitTime: int):
        """
        Call self.run after waitTime milliseconds.
        :param waitTime: milliseconds to wait for.
        """
        pass


class AsyncIOSequencer(Sequencer):
    """
    Sequencer that uses asyncio.sleep to wait between calls.
    """

    def wait(self, waitTime: int):
        async def waitFunction():
            await asyncio.sleep(waitTime / 1000.0)
            self.run()

        defer(waitFunction())


#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pathlib import Path
from typing import Optional
from pyrdp.core import settings
from pyrdp.enum import NegotiationProtocols


class MITMConfig:
    """
    Configuration options for the RDP MITM.
    """

    def __init__(self):
        self.targetHost: str = None
        """The RDP server"""

        self.targetPort: int = None
        """The RDP server's port"""

        self.listenAddress: str = "0.0.0.0"
        """The address to listen on."""

        self.listenPort: int = 3389
        """The port to bind for listening."""

        self.certificateFileName: str = None
        """Path to the TLS certificate"""

        self.privateKeyFileName: str = None
        """Path to the TLS private key"""

        self.attackerHost: Optional[str] = None
        """The attacker host"""

        self.attackerPort: Optional[int] = None
        """The attacker port"""

        self.replacementUsername: str = None
        """The replacement username for login attempts"""

        self.replacementPassword: str = None
        """The replacement password for login attempts"""

        self.outDir: Path = None
        """The output directory"""

        self.extractFiles: bool = True
        """Whether to extract file transferred between the client and server."""

        self.recordReplays: bool = True
        """Whether replays should be recorded or not"""

        self.downgrade: bool = True
        """Whether to actively downgrade unsupported extensions."""

        self.payload: str = ""
        """Payload to send automatically upon connection"""

        self.payloadDelay: int = None
        """Delay before sending payload automatically, in milliseconds"""

        self.payloadDuration: int = None
        """Amount of time the payload should take to complete, in milliseconds"""

        self.enableCrawler: bool = False
        """Whether the crawler should be enabled or not"""

        self.crawlerMatchFileName: str = None
        """Path to the crawler match configuration file"""

        self.crawlerIgnoreFileName: str = None
        """Path to the crawler ignore configuration file"""

        self.disableActiveClipboardStealing: bool = False
        """ If set to False, use PassiveClipboardStealer instead of ActiveClipboardStealer."""

        self.useGdi: bool = False
        """Whether to allow the client to use the GDI rendering pipeline extension."""

        self.authMethods: NegotiationProtocols = NegotiationProtocols.SSL
        """Specifies the list of authentication protocols that PyRDP accepts."""

        self.redirectionHost = None
        """Host to redirect the connection to if NLA is enforced"""

        self.redirectionPort = None
        """Port of the redirection host"""

    @property
    def replayDir(self) -> Path:
        """
        Get the directory for replay files.
        """
        return self.outDir / "replays"

    @property
    def fileDir(self) -> Path:
        """
        Get the directory for intercepted files.
        """
        return self.outDir / "files"

    @property
    def filesystemDir(self) -> Path:
        """
        Get the directory for filesystem clones.
        """
        return self.outDir / "filesystems"

    @property
    def certDir(self) -> Path:
        """
        Get the directory for dynamically generated certificates.
        """
        return self.outDir / "certs"


"""
The default MITM configuration.
"""
DEFAULTS = settings.load(Path(__file__).parent.absolute() / "mitm.default.ini")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pyrdp.layer.layer import IntermediateLayer
from pyrdp.parser import SegmentationParser
from pyrdp.pdu import PDU


class BufferedLayer(IntermediateLayer):
    """
    Base class for Layers that can receive PDUs spread across multiple calls to recv.
    """

    def __init__(self, parser: SegmentationParser):
        super().__init__(parser)
        self.buffer = b""

    def getDataLengthRequired(self) -> int:
        """
        Get the minimum amount that must be read next.
        This does not have to be the actual size of the PDU: if you return 1 twice in a row, you will read 1 byte twice
        in a row.
        This should always return 0 if the buffer is empty.
        """
        if self.buffer == b"":
            return 0

        try:
            pduLength = self.mainParser.getPDULength(self.buffer)
        except ValueError:
            return 1

        return pduLength - len(self.buffer)

    def recv(self, data: bytes):
        """
        Buffer data until we have a complete PDU, then parse the PDU and process it.
        :param data: received bytes.
        """
        data = self.buffer + data

        while len(data) > 0:
            if not self.mainParser.isCompletePDU(data):
                self.buffer = data
                data = b""
            else:
                pduLength = self.mainParser.getPDULength(data)
                pduData = data[: pduLength]
                data = data[pduLength :]
                self.buffer = b""

                pdu = self.mainParser.parse(pduData)
                self.pduReceived(pdu)

    def shouldForward(self, pdu: PDU) -> bool:
        return True



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
import typing

def defer(coroutine: typing.Union[typing.Coroutine, asyncio.Future]):
    """
    Create a twisted Deferred from a coroutine or future and ensure it will run (call ensureDeferred on it).
    :param coroutine: coroutine to defer.
    """
    from twisted.internet.defer import ensureDeferred, Deferred

    ensureDeferred(Deferred.fromFuture(asyncio.ensure_future(coroutine)))


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from typing import Callable

from pyrdp.core import ObservedBy
from pyrdp.enum import SlowPathPDUType, SlowPathDataType
from pyrdp.layer.layer import Layer, LayerStrictRoutedObserver
from pyrdp.parser import SlowPathParser
from pyrdp.pdu import ConfirmActivePDU, DemandActivePDU, SlowPathPDU


class SlowPathObserver(LayerStrictRoutedObserver):
    """
    Observer for slow-path PDUs.
    """

    def __init__(self, **kwargs):
        LayerStrictRoutedObserver.__init__(self, {
            SlowPathPDUType.DEMAND_ACTIVE_PDU: "onDemandActive",
            SlowPathPDUType.CONFIRM_ACTIVE_PDU: "onConfirmActive",
            SlowPathPDUType.DEACTIVATE_ALL_PDU: "onDeactivateAll",
            SlowPathPDUType.DATA_PDU: "onData",
            SlowPathPDUType.SERVER_REDIR_PKT_PDU: "onServerRedirect",
        }, **kwargs)

        self.dataHandlers = {}
        self.defaultDataHandler = None

    def setDataHandler(self, type: SlowPathDataType, handler: Callable[[SlowPathPDU], None]):
        """
        Set a handler for a particular data PDU type.
        :param type: PDU type for this handler.
        :param handler: the callback.
        """
        self.dataHandlers[type] = handler

    def setDefaultDataHandler(self, handler: Callable[[SlowPathPDU], None]):
        """
        Set the default handler.
        The default handler is called when a Data PDU is received that is not associated with a handler.
        """
        self.defaultDataHandler = handler

    def onPDUReceived(self, pdu: SlowPathPDU):
        if pdu.header.pduType in self.handlers:
            self.handlers[pdu.header.pduType](pdu)
        else:
            self.onUnknownHeader(pdu)

    def onData(self, pdu):
        """
        Called when a data PDU is received.
        :param pdu: the pdu.
        """
        self.dispatchPDU(pdu)

    def dispatchPDU(self, pdu: SlowPathPDU):
        """
        Call the proper handler depending on the PDU's subtype.
        :param pdu: the PDU that was received.
        """
        subtype = pdu.header.subtype

        if subtype in self.dataHandlers:
            self.dataHandlers[subtype](pdu)
        elif self.defaultDataHandler:
            self.defaultDataHandler(pdu)

    def onDemandActive(self, pdu: DemandActivePDU):
        """
        Called when a Demand Active PDU is received.
        Disable Virtual channel compression (unsupported for now).
        """
        pass

    def onConfirmActive(self, pdu: ConfirmActivePDU):
        """
        Change the received ConfirmActivePDU to facilitate data interception.
        """
        pass

    def onDeactivateAll(self, pdu):
        """
        Called when a Deactive All PDU is received.
        :param pdu: the PDU.
        """
        pass

    def onServerRedirect(self, pdu):
        """
        Called when a Server Redirect PDU is received.
        :param pdu: the PDU.
        """
        pass

@ObservedBy(SlowPathObserver)
class SlowPathLayer(Layer):
    """
    Layer for slow-path PDUs.
    """

    def __init__(self, parser = SlowPathParser()):
        Layer.__init__(self, parser)

    def sendBytes(self, data):
        self.previous.sendBytes(data)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import struct
import typing
from io import BytesIO

from Crypto.PublicKey import RSA
from Crypto.Util.number import bytes_to_long, long_to_bytes

from pyrdp.core import decodeUTF16LE, encodeUTF16LE, StrictStream, Uint16LE, Uint32LE, Uint8
from pyrdp.enum import ColorDepth, ConnectionDataType, ConnectionType, DesktopOrientation, EncryptionLevel, \
    EncryptionMethod, HighColorDepth, RDPVersion, ServerCertificateType
from pyrdp.exceptions import ParsingError, UnknownPDUTypeError, ExploitError
from pyrdp.parser.parser import Parser
from pyrdp.pdu import ClientChannelDefinition, ClientClusterData, ClientCoreData, ClientDataPDU, ClientNetworkData, \
    ClientSecurityData, ProprietaryCertificate, ServerCoreData, ServerDataPDU, ServerNetworkData, ServerSecurityData


class ClientConnectionParser(Parser):
    """
    Parser for Client Data PDUs (i.e: servers).
    """
    def __init__(self):
        super().__init__()

        self.parsers = {
            ConnectionDataType.CLIENT_CORE: self.parseClientCoreData,
            ConnectionDataType.CLIENT_SECURITY: self.parseClientSecurityData,
            ConnectionDataType.CLIENT_NETWORK: self.parseClientNetworkData,
            ConnectionDataType.CLIENT_CLUSTER: self.parseClientClusterData,
        }

        self.writers = {
            ConnectionDataType.CLIENT_CORE: self.writeClientCoreData,
            ConnectionDataType.CLIENT_SECURITY: self.writeClientSecurityData,
            ConnectionDataType.CLIENT_NETWORK: self.writeClientNetworkData,
            ConnectionDataType.CLIENT_CLUSTER: self.writeClientClusterData,

        }

    def doParse(self, data: bytes) -> ClientDataPDU:
        """
        Decode a Client Data PDU from bytes.
        :param data: Client Data PDU data.
        """
        core = None
        security = None
        network = None
        cluster = None

        stream = BytesIO(data)
        while stream.tell() != len(stream.getvalue()) and (core is None or security is None or network is None or cluster is None):
            structure = self.parseStructure(stream)

            if structure.header == ConnectionDataType.CLIENT_CORE:
                core = structure
            elif structure.header == ConnectionDataType.CLIENT_SECURITY:
                security = structure
            elif structure.header == ConnectionDataType.CLIENT_NETWORK:
                network = structure
            elif structure.header == ConnectionDataType.CLIENT_CLUSTER:
                cluster = structure

            if len(stream.getvalue()) == 0:
                break

        return ClientDataPDU(core, security, network, cluster)

    def parseStructure(self, stream: BytesIO) -> typing.Union[ClientCoreData, ClientNetworkData, ClientSecurityData, ClientClusterData]:
        header = Uint16LE.unpack(stream)
        length = Uint16LE.unpack(stream) - 4
        data = stream.read(length)

        if len(data) != length:
            raise ParsingError("Client Data length field does not match actual size")

        substream = BytesIO(data)

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown client data structure %s" % header, header)

        return self.parsers[header](substream)

    def parseClientCoreData(self, stream: BytesIO) -> ClientCoreData:
        stream = StrictStream(stream)

        # 128 bytes minimum (excluding header)
        version = RDPVersion(Uint32LE.unpack(stream))
        desktopWidth = Uint16LE.unpack(stream)
        desktopHeight = Uint16LE.unpack(stream)
        colorDepth = ColorDepth(Uint16LE.unpack(stream))
        sasSequence = Uint16LE.unpack(stream)
        keyboardLayout = Uint32LE.unpack(stream)
        clientBuild = Uint32LE.unpack(stream)
        clientName = decodeUTF16LE(stream.read(32))
        keyboardType = Uint32LE.unpack(stream)
        keyboardSubType = Uint32LE.unpack(stream)
        keyboardFunctionKey = Uint32LE.unpack(stream)
        imeFileName = stream.read(64)

        core = ClientCoreData(version, desktopWidth, desktopHeight, colorDepth, sasSequence, keyboardLayout, clientBuild, clientName, keyboardType, keyboardSubType, keyboardFunctionKey, imeFileName)

        # Optional data
        # The optional fields are read in order. If one of them is not present, then all subsequent fields are also not present.
        try:
            core.postBeta2ColorDepth = Uint16LE.unpack(stream)
            core.clientProductId = Uint16LE.unpack(stream)
            core.serialNumber = Uint32LE.unpack(stream)

            # Should match HighColorDepth enum most of the time, but in order to support scanners and we script, we have to loosely accept this one
            # Anyway, the server will reject it and enforce another one
            core.highColorDepth = Uint16LE.unpack(stream)
            core.supportedColorDepths = Uint16LE.unpack(stream)
            core.earlyCapabilityFlags = Uint16LE.unpack(stream)
            core.clientDigProductId = decodeUTF16LE(stream.read(64))
            core.connectionType = ConnectionType(Uint8.unpack(stream))
            stream.read(1)
            core.serverSelectedProtocol = Uint32LE.unpack(stream)
            core.desktopPhysicalWidth = Uint32LE.unpack(stream)
            core.desktopPhysicalHeight = Uint32LE.unpack(stream)
            core.desktopOrientation = DesktopOrientation(Uint16LE.unpack(stream))
            core.desktopScaleFactor = Uint32LE.unpack(stream)
            core.deviceScaleFactor = Uint32LE.unpack(stream)
        except EOFError:
            # The stream has reached the end, we don't have any more optional fields. This exception can be ignored.
            pass

        return core

    def parseClientSecurityData(self, stream: BytesIO) -> ClientSecurityData:
        encryptionMethods = Uint32LE.unpack(stream)
        extEncryptionMethods = Uint32LE.unpack(stream)
        return ClientSecurityData(encryptionMethods, extEncryptionMethods)

    def detectBlueKeepScan(self, data: bytes):
        return b"MS_T120" in data.upper()

    def parseClientNetworkData(self, stream: BytesIO) -> ClientNetworkData:
        if self.detectBlueKeepScan(stream.getvalue()):
            raise ExploitError("BlueKeep scan or exploit attempted")

        channelCount = Uint32LE.unpack(stream)
        data = stream.getvalue()[4 :]

        if len(data) != channelCount * 12:
            raise ParsingError("Invalid channel array size")

        channelDefinitions = []

        for _ in range(channelCount):
            name = stream.read(8).strip(b"\x00").decode()
            options = Uint32LE.unpack(stream)
            channelDefinitions.append(ClientChannelDefinition(name, options))

        return ClientNetworkData(channelDefinitions)

    def parseClientClusterData(self, stream: BytesIO) -> ClientClusterData:
        flags = Uint32LE.unpack(stream)
        redirectedSessionID = Uint32LE.unpack(stream)
        return ClientClusterData(flags, redirectedSessionID)

    def write(self, pdu: ClientDataPDU) -> bytes:
        """
        Encode a Client Data PDU to bytes.
        :param pdu: the Client Data PDU
        """
        stream = BytesIO()

        if pdu.coreData:
            self.writeStructure(stream, pdu.coreData)

        if pdu.securityData:
            self.writeStructure(stream, pdu.securityData)

        if pdu.networkData:
            self.writeStructure(stream, pdu.networkData)

        if pdu.clusterData:
            self.writeStructure(stream, pdu.clusterData)

        return stream.getvalue()

    def writeStructure(self, stream: BytesIO, data: typing.Union[ClientCoreData, ClientNetworkData, ClientSecurityData, ClientClusterData]):
        if data.header not in self.writers:
            raise UnknownPDUTypeError("Trying to write unknown Client Data structure %s" % data.header, data.header)

        substream = BytesIO()
        self.writers[data.header](substream, data)

        substream = substream.getvalue()

        stream.write(Uint16LE.pack(data.header))
        stream.write(Uint16LE.pack(len(substream) + 4))
        stream.write(substream)

    def writeClientCoreData(self, stream: BytesIO, core: ClientCoreData):
        stream.write(Uint32LE.pack(core.version))
        stream.write(Uint16LE.pack(core.desktopWidth))
        stream.write(Uint16LE.pack(core.desktopHeight))
        stream.write(Uint16LE.pack(core.colorDepth))
        stream.write(Uint16LE.pack(core.sasSequence))
        stream.write(Uint32LE.pack(core.keyboardLayout))
        stream.write(Uint32LE.pack(core.clientBuild))
        stream.write(encodeUTF16LE(core.clientName).ljust(32, b"\x00")[: 32])
        stream.write(Uint32LE.pack(core.keyboardType))
        stream.write(Uint32LE.pack(core.keyboardSubType))
        stream.write(Uint32LE.pack(core.keyboardFunctionKey))
        stream.write(core.imeFileName)

        try:
            stream.write(Uint16LE.pack(core.postBeta2ColorDepth))
            stream.write(Uint16LE.pack(core.clientProductId))
            stream.write(Uint32LE.pack(core.serialNumber))
            stream.write(Uint16LE.pack(core.highColorDepth))
            stream.write(Uint16LE.pack(core.supportedColorDepths))
            stream.write(Uint16LE.pack(core.earlyCapabilityFlags))
            stream.write(encodeUTF16LE(core.clientDigProductId).ljust(64, b"\x00")[: 64])
            stream.write(Uint8.pack(core.connectionType))
            stream.write(b"\x00")
            stream.write(Uint32LE.pack(core.serverSelectedProtocol))
            stream.write(Uint32LE.pack(core.desktopPhysicalWidth))
            stream.write(Uint32LE.pack(core.desktopPhysicalHeight))
            stream.write(Uint16LE.pack(core.desktopOrientation))
            stream.write(Uint32LE.pack(core.desktopScaleFactor))
            stream.write(Uint32LE.pack(core.deviceScaleFactor))
        except struct.error:
            # We tried to write an optional field which was not present. Stop writing beyond this point.
            pass

    def writeClientSecurityData(self, stream: BytesIO, security: ClientSecurityData):
        stream.write(Uint32LE.pack(security.encryptionMethods))
        stream.write(Uint32LE.pack(security.extEncryptionMethods))

    def writeClientNetworkData(self, stream: BytesIO, network: ClientNetworkData):
        stream.write(Uint32LE.pack(len(network.channelDefinitions)))

        for channel in network.channelDefinitions:
            if len(channel.name) > 8:
                raise ParsingError("Channel name must have 8 characters maximum")

            stream.write(channel.name.encode().ljust(8, b"\x00")[: 8])
            stream.write(Uint32LE.pack(channel.options))

    def writeClientClusterData(self, stream: BytesIO, cluster: ClientClusterData):
        stream.write(Uint32LE.pack(cluster.flags))
        stream.write(Uint32LE.pack(cluster.redirectedSessionID))


class ServerConnectionParser(Parser):
    """
    Parser for Server Data PDUs (i.e: client).
    """

    def __init__(self):
        super().__init__()
        self.parsers = {
            ConnectionDataType.SERVER_CORE: self.parseServerCoreData,
            ConnectionDataType.SERVER_NETWORK: self.parseServerNetworkData,
            ConnectionDataType.SERVER_SECURITY: self.parseServerSecurityData,
        }

        self.writers = {
            ConnectionDataType.SERVER_CORE: self.writeServerCoreData,
            ConnectionDataType.SERVER_NETWORK: self.writeServerNetworkData,
            ConnectionDataType.SERVER_SECURITY: self.writeServerSecurityData,
        }

    def doParse(self, data: bytes) -> ServerDataPDU:
        """
        Parse a Server Data PDU from bytes.
        """
        core = None
        security = None
        network = None

        stream = BytesIO(data)
        while core is None or security is None or network is None:
            structure = self.parseStructure(stream)

            if structure.header == ConnectionDataType.SERVER_CORE:
                core = structure
            elif structure.header == ConnectionDataType.SERVER_SECURITY:
                security = structure
            elif structure.header == ConnectionDataType.SERVER_NETWORK:
                network = structure

            if len(stream.getvalue()) == 0:
                break

        return ServerDataPDU(core, security, network)

    def parseStructure(self, stream: BytesIO) -> typing.Union[ServerCoreData, ServerSecurityData, ServerNetworkData]:
        header = Uint16LE.unpack(stream)
        length = Uint16LE.unpack(stream) - 4
        data = stream.read(length)

        if len(data) < length:
            raise ParsingError("Server Data length field does not match actual size")

        substream = BytesIO(data)

        if header not in self.parsers:
            raise UnknownPDUTypeError("Trying to parse unknown server data structure %s" % header, header)

        return self.parsers[header](substream)

    def parseServerCoreData(self, stream: BytesIO) -> ServerCoreData:
        stream = StrictStream(stream)

        clientRequestedProtocols = None
        earlyCapabilityFlags = None
        version = Uint32LE.unpack(stream)

        try:
            clientRequestedProtocols = Uint32LE.unpack(stream)
            earlyCapabilityFlags = Uint32LE.unpack(stream)
        except EOFError:
            pass

        return ServerCoreData(version, clientRequestedProtocols, earlyCapabilityFlags)

    def parseServerNetworkData(self, stream: BytesIO) -> ServerNetworkData:
        mcsChannelID = Uint16LE.unpack(stream)
        channelCount = Uint16LE.unpack(stream)
        channels = [Uint16LE.unpack(stream) for _ in range(channelCount)]

        return ServerNetworkData(mcsChannelID, channels)

    def parseServerSecurityData(self, stream: BytesIO) -> ServerSecurityData:
        stream = StrictStream(stream)
        encryptionMethod = EncryptionMethod(Uint32LE.unpack(stream))
        encryptionLevel = EncryptionLevel(Uint32LE.unpack(stream))
        serverRandom = None
        serverCertificate = None

        try:
            serverRandomLength = Uint32LE.unpack(stream)
            serverCertificateLength = Uint32LE.unpack(stream)
            serverRandom = stream.read(serverRandomLength)
            serverCertificate = stream.read(serverCertificateLength)
            serverCertificate = self.parseServerCertificate(serverCertificate)
        except EOFError:
            pass

        return ServerSecurityData(encryptionMethod, encryptionLevel, serverRandom, serverCertificate)

    def parseServerCertificate(self, data: bytes) -> ProprietaryCertificate:
        stream = BytesIO(data)
        version = Uint32LE.unpack(stream)

        if version == ServerCertificateType.PROPRIETARY:
            return self.parseProprietaryCertificate(stream)
        else:
            raise NotImplementedError("Unhandled certificate type")

    def parseProprietaryCertificate(self, stream: BytesIO) -> ProprietaryCertificate:
        signatureAlgorithmID = Uint32LE.unpack(stream)
        keyAlgorithmID = Uint32LE.unpack(stream)
        publicKeyType = Uint16LE.unpack(stream)
        keyLength = Uint16LE.unpack(stream)
        publicKey = stream.read(keyLength)
        signatureType = Uint16LE.unpack(stream)
        signatureLength = Uint16LE.unpack(stream)
        signature = stream.read(signatureLength - 8)
        padding = stream.read()

        publicKey = self.parsePublicKey(publicKey)

        return ProprietaryCertificate(signatureAlgorithmID, keyAlgorithmID, publicKeyType, publicKey, signatureType, signature, padding)

    def parsePublicKey(self, data: bytes) -> RSA.RsaKey:
        stream = BytesIO(data)
        _magic = stream.read(4)
        keyLength = Uint32LE.unpack(stream)
        _bitLength = Uint32LE.unpack(stream)
        _dataLength = Uint32LE.unpack(stream)
        publicExponent = Uint32LE.unpack(stream)
        modulus = stream.read(keyLength - 8)
        _padding = stream.read(8)

        # Modulus must be reversed because bytes_to_long expects it to be in big endian format
        modulus = bytes_to_long(modulus[:: -1])
        publicExponent = int(publicExponent)
        publicKey = RSA.construct((modulus, publicExponent))
        return publicKey

    def write(self, pdu: ServerDataPDU) -> bytes:
        """
        Encode a Server Data PDU to bytes
        :param pdu: the Server Data PDU
        """
        stream = BytesIO()

        if pdu.coreData:
            self.writeStructure(stream, pdu.coreData)

        if pdu.securityData:
            self.writeStructure(stream, pdu.securityData)

        if pdu.networkData:
            self.writeStructure(stream, pdu.networkData)

        return stream.getvalue()

    def writeStructure(self, stream: BytesIO, data: typing.Union[ServerCoreData, ServerSecurityData, ServerNetworkData]):
        """
        :param stream: BytesIO to write to
        :param data: The structure to write (ex: ServerCoreData)
        """
        if data.header not in self.writers:
            raise UnknownPDUTypeError("Trying to write unknown Server Data structure %s" % data.header, data.header)

        substream = BytesIO()
        self.writers[data.header](substream, data)

        substream = substream.getvalue()

        stream.write(Uint16LE.pack(data.header))
        stream.write(Uint16LE.pack(len(substream) + 4))
        stream.write(substream)

    def writeServerCoreData(self, stream: BytesIO, data: ServerCoreData):
        stream.write(Uint32LE.pack(data.version))

        requestedProtocols = data.clientRequestedProtocols

        if requestedProtocols is None:
            requestedProtocols = 0

        stream.write(Uint32LE.pack(requestedProtocols))

        if data.earlyCapabilityFlags is not None:
            stream.write(Uint32LE.pack(data.earlyCapabilityFlags))

    def writeServerNetworkData(self, stream: BytesIO, data: ServerNetworkData):
        """
        https://msdn.microsoft.com/en-us/library/cc240522.aspx
        """
        stream.write(Uint16LE.pack(data.mcsChannelID))
        stream.write(Uint16LE.pack(len(data.channels)))

        for channel in data.channels:
            stream.write(Uint16LE.pack(channel))

        if len(data.channels) % 2 != 0:
            stream.write(Uint16LE.pack(0))  # Write 2 empty bytes so we keep a multiple of 4.

    def writeServerSecurityData(self, stream: BytesIO, data: ServerSecurityData):
        stream.write(Uint32LE.pack(data.encryptionMethod))
        stream.write(Uint32LE.pack(data.encryptionLevel))
        if data.serverRandom is not None:
            serverCertificate = self.writeServerCertificate(data.serverCertificate)

            stream.write(Uint32LE.pack(len(data.serverRandom)))
            stream.write(Uint32LE.pack(len(serverCertificate)))
            stream.write(data.serverRandom)
            stream.write(serverCertificate)

    def writeServerCertificate(self, certificate: ProprietaryCertificate) -> bytes:
        stream = BytesIO()

        if certificate.type == ServerCertificateType.PROPRIETARY:
            Uint32LE.pack(ServerCertificateType.PROPRIETARY, stream)
            self.writeProprietaryCertificate(stream, certificate)
        else:
            raise NotImplementedError("Unhandled certificate type")

        return stream.getvalue()

    def writeProprietaryCertificate(self, stream: BytesIO, cert: ProprietaryCertificate):
        keyBytes = self.writePublicKey(cert.publicKey)

        Uint32LE.pack(cert.signatureAlgorithmID, stream)
        Uint32LE.pack(cert.keyAlgorithmID, stream)
        Uint16LE.pack(cert.publicKeyType, stream)
        Uint16LE.pack(len(keyBytes), stream)
        stream.write(keyBytes)
        Uint16LE.pack(cert.signatureType, stream)
        Uint16LE.pack(len(cert.signature) + 8, stream)
        stream.write(cert.signature)
        stream.write(b"\x00" * 8)

    def writePublicKey(self, publicKey: RSA.RsaKey) -> bytes:
        modulus = publicKey.n
        publicExponent = publicKey.e

        # Modulus must be reversed because bytes_to_long expects it to be in big endian format
        modulusBytes = long_to_bytes(modulus)[:: -1]

        stream = BytesIO()
        stream.write(b"RSA1")
        Uint32LE.pack(len(modulusBytes) + 8, stream)
        Uint32LE.pack(2048, stream)
        Uint32LE.pack(255, stream)
        Uint32LE.pack(publicExponent, stream)
        stream.write(modulusBytes)
        stream.write(b"\x00" * 8)
        return stream.getvalue()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging
from binascii import hexlify

from pyrdp.security import SecuritySettingsObserver


class RC4LoggingObserver(SecuritySettingsObserver):
    def __init__(self, log: logging.LoggerAdapter):
        super().__init__()
        self.log = log

    def onCrypterGenerated(self, settings):
        self.log.debug("RC4 client/server random: %(rc4ClientRandom)s %(rc4ServerRandom)s",
                       {"rc4ClientRandom": hexlify(settings.clientRandom).decode(),
                        "rc4ServerRandom": hexlify(settings.serverRandom).decode()})


#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import LoggerAdapter
from io import BytesIO
from functools import partial

from pathlib import Path

from pyrdp.core import decodeUTF16LE, Uint64LE
from pyrdp.enum import ClipboardFormatNumber, ClipboardMessageFlags, ClipboardMessageType, PlayerPDUType, FileContentsFlags
from pyrdp.layer import ClipboardLayer
from pyrdp.logging.StatCounter import StatCounter, STAT
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu import ClipboardPDU, FormatDataRequestPDU, FormatDataResponsePDU, FileContentsRequestPDU, FileContentsResponsePDU
from pyrdp.parser.rdp.virtual_channel.clipboard import FileDescriptor
from pyrdp.recording import Recorder
from pyrdp.mitm.config import MITMConfig
from pyrdp.mitm.FileMapping import FileMapping

from twisted.internet import reactor  # Import the current reactor.
from twisted.python.failure import Failure


TRANSFER_TIMEOUT = 5  # delay in seconds after which to kill a stalled transfer.
CLIPBOARD_FILEDIR = "clipboard" # special directory name under filesystems/<sessionId> for collected clipboard files


class PassiveClipboardStealer:
    """
    MITM component for the clipboard layer. Logs clipboard data when it is pasted.
    """

    def __init__(self, config: MITMConfig, client: ClipboardLayer, server: ClipboardLayer, log: LoggerAdapter, recorder: Recorder,
                 statCounter: StatCounter, state: RDPMITMState):
        """
        :param client: clipboard layer for the client side
        :param server: clipboard layer for the server side
        :param log: logger for this component
        :param recorder: recorder for clipboard data
        """
        self.statCounter = statCounter
        self.client = client
        self.server = server
        self.config = config
        self.log = log
        self.state = state
        self.recorder = recorder
        self.forwardNextDataResponse = True
        self.files = []
        self.transfers = {}
        self.timeouts = {}  # Track active timeout monitoring tasks.

        self.filesystemRoot = self.config.filesystemDir / self.state.sessionID

        self.client.createObserver(
            onPDUReceived = self.onClientPDUReceived,
        )

        self.server.createObserver(
            onPDUReceived = self.onServerPDUReceived,
        )

        # Dispatchers must return whether to forward the packet.
        self.dispatch = {
            FormatDataResponsePDU: self.onFormatDataResponse,
        }

        # Only handle file contents if file extraction is enabled.
        if self.config.extractFiles:
            self.dispatch[FileContentsRequestPDU] = self.onFileContentsRequest
            self.dispatch[FileContentsResponsePDU] = self.onFileContentsResponse

    def onClientPDUReceived(self, pdu: ClipboardPDU):
        self.statCounter.increment(STAT.CLIPBOARD, STAT.CLIPBOARD_CLIENT)
        self.handlePDU(pdu, self.server)

    def onServerPDUReceived(self, pdu: ClipboardPDU):
        self.statCounter.increment(STAT.CLIPBOARD, STAT.CLIPBOARD_SERVER)
        self.handlePDU(pdu, self.client)

    def handlePDU(self, pdu: ClipboardPDU, destination: ClipboardLayer):
        """
        Handle an incoming clipboard message.

        :param pdu: the PDU that was received
        :param destination: the destination layer
        """

        forward = True
        # Handle file transfers
        if type(pdu) in self.dispatch:
            forward = self.dispatch[type(pdu)](pdu)
        assert forward is not None, "ClipboardMITM: PDU handler must return True or False!"

        if forward:
            destination.sendPDU(pdu)

    def onFileContentsRequest(self, pdu: FileContentsRequestPDU):
        """
        There are two types of content requests: SIZE and RANGE.

        A new transfer begins with a SIZE request and is followed by multiple
        RANGE requests to retrieve the file data.

        The file is picked from the advertised clipboard file list with an index.
        """
        if pdu.flags == FileContentsFlags.SIZE:
            if pdu.lindex < len(self.files):
                fd = self.files[pdu.lindex]
                self.log.info("Starting transfer for file '%(filename)s', clipId=%(clipId)d",
                              {"filename": fd.filename, "clipId": pdu.clipId})

                if pdu.streamId in self.transfers:
                    self.log.warning("Clipboard file transfer already started file '%(filename)s', clipId=%(clipId)d",
                                     {"filename": fd.filename, "clipId": pdu.clipId})

                self.transfers[pdu.streamId] = FileTransferMappingProxy(fd, self.config.fileDir, self.filesystemRoot, self.log, pdu.size)

                # Track transfer timeout to prevent hung transfers.
                cbTimeout = reactor.callLater(TRANSFER_TIMEOUT, partial(self.onTransferTimedOut, pdu.streamId))
                self.timeouts[pdu.streamId] = cbTimeout
            else:
                self.log.info('Request for unknown file! (list index=%(lIndex)d)',
                              {'lIndex': pdu.lindex})

        elif pdu.flags == FileContentsFlags.RANGE:
            if pdu.streamId not in self.transfers:
                self.log.warning('FileContentsRequest for unknown transfer (streamId=%(streamId)d)',
                                 {"streamId": pdu.streamId})
            else:
                self.refreshTimeout(pdu.streamId)
                self.transfers[pdu.streamId].onRequest(pdu)

        return True

    def onFileContentsResponse(self, pdu: FileContentsResponsePDU):
        if pdu.streamId not in self.transfers:
            self.log.warning('FileContentsResponse for unknown transfer (streamId=%(streamId)d)',
                             {"streamId": pdu.streamId})
        else:
            self.refreshTimeout(pdu.streamId)

            done = self.transfers[pdu.streamId].onResponse(pdu)
            if done:
                xfer = self.transfers[pdu.streamId]
                self.log.info("Clipboard transfer completed for file '%(filename)s', saved as: '%(localPath)s', "
                              "linked from: '%(linkPath)s'",
                              {"filename": xfer.info.filename, "localPath": str(self.config.fileDir / xfer.getFileHash()),
                               "linkPath": str(xfer.getFilesystemPath())})
                del self.transfers[pdu.streamId]

                # Remove the timeout since the transfer is done.
                # This cannot throw because if we got this far, the delayed task cannot
                # have been executed yet.
                self.timeouts[pdu.streamId].cancel()
                del self.timeouts[pdu.streamId]

        return True

    def onTransferTimedOut(self, streamId: int):
        if streamId in self.transfers:
            # If the transfer exists, abort it. Otherwise, most likely the
            # transfer has been completed. The latter should never happen due to the way
            # twisted's reactor works.
            xfer = self.transfers[streamId]
            xfer.onDisconnection(Failure(Exception("Clipboard transfer timeout")))
            self.log.warn("Transfer timed out for '%(filename)s' saved to: '%(localPath)s'",
                          {"filename": xfer.info.filename, "localPath": str(xfer.getDataPath())})
            del self.transfers[streamId]
            del self.timeouts[streamId]

    def refreshTimeout(self, streamId: int):
        self.timeouts[streamId].delay(TRANSFER_TIMEOUT)

    def onFormatDataResponse(self, pdu: FormatDataResponsePDU):
        if pdu.msgFlags == ClipboardMessageFlags.CB_RESPONSE_OK:
            # Keep the file list if there is one.
            # FIXME: There is currently no concept of transfer direction.
            if len(pdu.files) > 0:
                self.files = pdu.files

            if pdu.formatId == ClipboardFormatNumber.GENERIC.value:
                clipboardData = self.decodeClipboardData(pdu.requestedFormatData)
                self.log.info("Clipboard data: %(clipboardData)r", {"clipboardData": clipboardData})
                # FIXME: Record all clipboard related messages?
                self.recorder.record(pdu, PlayerPDUType.CLIPBOARD_DATA)

            if self.forwardNextDataResponse:
                # Means it's NOT a crafted response
                self.statCounter.increment(STAT.CLIPBOARD_PASTE)

        # Do not forward the response if it is for an injected DataRequest.
        forward = self.forwardNextDataResponse
        self.forwardNextDataResponse = True
        return forward

    def decodeClipboardData(self, data: bytes) -> str:
        """
        Decode clipboard bytes to a string.
        :param data: clipboard content bytes
        """
        return decodeUTF16LE(data)


class ActiveClipboardStealer(PassiveClipboardStealer):
    """
    MITM component for the clipboard layer. Actively spies on the clipboard by sending paste requests whenever the
    clipboard is updated.
    """

    def __init__(self, config: MITMConfig, client: ClipboardLayer, server: ClipboardLayer, log: LoggerAdapter, recorder: Recorder,
                 statCounter: StatCounter, state: RDPMITMState):
        super().__init__(config, client, server, log, recorder, statCounter, state)

    def handlePDU(self, pdu: ClipboardPDU, destination: ClipboardLayer):
        """
        Check if the PDU is a FormatListResponse. If so, send a paste request to steal the clipboard data.
        :param pdu: the PDU that was received
        :param destination: the destination layer
        """
        super().handlePDU(pdu, destination)

        if pdu.msgType == ClipboardMessageType.CB_FORMAT_LIST_RESPONSE:
            self.sendPasteRequest(destination)

    def sendPasteRequest(self, destination: ClipboardLayer):
        """
        Send a FormatDataRequest to request the clipboard data.
        Sets forwardNextDataResponse to False to make sure that this request is not actually transferred to the other end.
        """

        self.statCounter.increment(STAT.CLIPBOARD_COPY)

        formatDataRequestPDU = FormatDataRequestPDU(ClipboardFormatNumber.GENERIC)
        destination.sendPDU(formatDataRequestPDU)
        self.forwardNextDataResponse = False

class FileTransferMappingProxy():
    """Encapsulate the state of a clipboard file transfer but proxies to FileMapping for storage and logging consistency"""
    def __init__(self, fd: FileDescriptor, outDir: Path, filesystemSessionIdRoot: Path, log: LoggerAdapter, size: int):
        self.info = fd
        self.size = size
        self.transferred: int = 0
        self.prev = None  # Pending file content request.

        # We store files under filesystems/<sessionID>/ under a special clipboard directory
        symlinkDst = filesystemSessionIdRoot / CLIPBOARD_FILEDIR
        symlinkDst.mkdir(parents=True, exist_ok=True)

        self.fileMapping = FileMapping.generate("/" + fd.filename, outDir, symlinkDst, log)

    def onRequest(self, pdu: FileContentsRequestPDU):
        # TODO: Handle out of order ranges. Are they even possible?
        self.prev = pdu

    def onResponse(self, pdu: FileContentsResponsePDU) -> bool:
        """
        Handle file data.

        @Returns True if file transfer is complete.
        """
        if not self.prev:
            # First response always contains file size.
            self.size = Uint64LE.unpack(BytesIO(pdu.data))

            return False

        received = len(pdu.data)

        self.fileMapping.write(pdu.data)
        self.transferred += received

        if self.transferred == self.size:
            self.fileMapping.finalize()
            return True

        return False

    def getFileHash(self) -> str:
        return self.fileMapping.fileHash

    def getFilesystemPath(self) -> Path:
        return self.fileMapping.filesystemPath

    def onDisconnection(self, reason):
        return self.fileMapping.onDisconnection(reason)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2024 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import datetime
import logging

from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QLabel, QSizePolicy, QSlider, QSpacerItem, \
    QVBoxLayout, QWidget

from pyrdp.logging import LOGGER_NAMES
from pyrdp.player.SeekBar import SeekBar
from pyrdp.ui import PlayPauseButton


class ReplayBar(QWidget):
    """
    Widget that contains the play/pause button, the seek bar and the speed slider.
    """
    play = Signal()
    pause = Signal()
    seek = Signal(float)
    speedChanged = Signal(int)

    def __init__(self, duration: float, referenceTime: int, parent: QWidget = None):
        super().__init__(parent)

        self.log = logging.getLogger(LOGGER_NAMES.PLAYER)
        
        # absolute capture starting point
        self.referenceTime = referenceTime
        
        # pretty display capture duration
        self.duration = duration
        self.duration_hours = int(self.duration // 3600)
        remaining_seconds = int(self.duration % 3600)
        minutes = int(remaining_seconds // 60)
        seconds = int(remaining_seconds % 60)
        if self.duration_hours > 0:
            self.duration_pretty = "{:d}:{:02d}:{:02d}".format(
                                    self.duration_hours, minutes, seconds)
        else:
            self.duration_pretty = "{:02d}:{:02d}".format(minutes, seconds)

        self.button = PlayPauseButton()
        self.button.setMaximumWidth(100)
        self.button.clicked.connect(self.onButtonClicked)

        self.scaleCheckbox = QCheckBox("Scale to window")

        self.timeSlider = SeekBar()
        self.timeSlider.setMinimum(0)
        self.timeSlider.setMaximum(int(self.duration * 1000))
        self.timeSlider.valueChanged.connect(self.onSeek)

        self.timeLabel = QLabel(self.formatTimeLabel(0))
        self.absoluteDateTimeLabel = QLabel(self.formatAbsoluteDateTimeLabel(0))

        self.speedLabel = QLabel("Speed: 1x")

        self.speedSlider = QSlider(Qt.Horizontal)
        self.speedSlider.setMaximumWidth(300)
        self.speedSlider.setMinimum(1)
        self.speedSlider.setMaximum(10)
        self.speedSlider.valueChanged.connect(self.onSpeedChanged)

        vertical = QVBoxLayout()

        horizontal = QHBoxLayout()
        horizontal.addWidget(self.speedLabel)
        horizontal.addWidget(self.speedSlider)
        horizontal.addWidget(self.scaleCheckbox)
        horizontal.addItem(QSpacerItem(20, 40, QSizePolicy.Expanding, QSizePolicy.Expanding))
        vertical.addLayout(horizontal)

        horizontal = QHBoxLayout()
        horizontal.addWidget(self.button)
        horizontal.addWidget(self.timeSlider)
        horizontal.addWidget(self.timeLabel)
        horizontal.addWidget(self.absoluteDateTimeLabel)
        vertical.addLayout(horizontal)

        self.setLayout(vertical)
        self.setGeometry(0, 0, 80, 60)

    def onButtonClicked(self):
        if self.button.playing:
            self.log.debug("Play clicked")
            self.play.emit()
        else:
            self.log.debug("Pause clicked")
            self.pause.emit()

    def onSeek(self):
        time = self.timeSlider.value() / 1000.0
        self.log.debug("Seek to %(arg1)d seconds", {"arg1": time})
        self.seek.emit(time)

    def onSpeedChanged(self):
        speed = self.speedSlider.value()
        self.log.debug("Slider changed value: %(arg1)d", {"arg1": speed})
        self.speedLabel.setText("Speed: {}x".format(speed))
        self.speedChanged.emit(speed)

    def onTimeChanged(self, currentTime: float):
        if currentTime >= self.duration:
            currentTime = self.duration
        
        self.timeLabel.setText(self.formatTimeLabel(currentTime))
        self.absoluteDateTimeLabel.setText(self.formatAbsoluteDateTimeLabel(currentTime))    

    def formatTimeLabel(self, current: float) -> str:
        hours = int(current // 3600)
        remaining_seconds = int(current % 3600)
        minutes = int(remaining_seconds // 60)
        seconds = int(remaining_seconds % 60)
        if self.duration_hours > 0:
            return "{:d}:{:02d}:{:02d} / {}".format(
                hours, minutes, seconds, self.duration_pretty)
        else:
            return "{:02d}:{:02d} / {}".format(minutes, seconds, self.duration_pretty)
        
    def formatAbsoluteDateTimeLabel(self, current: int) -> str:
        current = current * 1000
        epoch_time = (self.referenceTime + current)
        formatted_time = datetime.datetime.fromtimestamp(epoch_time / 1000, tz=datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S UTC')

        return formatted_time



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.ui.PlayPauseButton import PlayPauseButton
from pyrdp.ui.qt import QRemoteDesktop, RDPBitmapToQtImage



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum.virtual_channel.dynamic_channel import CbId, DynamicChannelCommand
from pyrdp.pdu import PDU


class DynamicChannelPDU(PDU):
    """
    Base for DynamicChannelPDUs
    https://msdn.microsoft.com/en-us/library/cc241267.aspx
    """

    def __init__(self, cbid: int, sp: int, cmd: int, payload=b""):
        super().__init__(payload)
        self.cbid = CbId(cbid)
        self.sp = sp
        self.cmd = DynamicChannelCommand(cmd)


class CreateRequestPDU(DynamicChannelPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241244.aspx
    """

    def __init__(self, cbid: int, sp: int, channelId: int, channelName: str):
        super().__init__(cbid, sp, DynamicChannelCommand.CREATE)
        self.channelId = channelId
        self.channelName = channelName


class CreateResponsePDU(DynamicChannelPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc241245.aspx
    """

    def __init__(self, cbid, sp, channelId: int, creationStatus: int):
        super().__init__(cbid, sp, DynamicChannelCommand.CREATE)
        self.channelId = channelId
        self.creationStatus = creationStatus



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.pdu.pdu import PDU


class SecurityPDU(PDU):
    def __init__(self, header, payload):
        PDU.__init__(self, payload)
        self.header = header


class SecurityExchangePDU(PDU):
    def __init__(self, header, clientRandom):
        super().__init__()
        self.header = header
        self.clientRandom = clientRandom
